"""
Module which defines the base controls for awsc. Ideally, all resource specific controls should inherit these and be created with as little
effort as possible.
"""
import datetime
import json
import threading
import traceback
from operator import attrgetter
from typing import Any, Callable, Dict, List

import pyperclip
from botocore import exceptions as botoerror

from .common import (
    Common,
    SessionAwareDialog,
    datetime_hack,
    default_args,
    default_kwargs,
)
from .hotkey import HotkeyDisplay
from .termui.alignment import CenterAnchor, Dimension
from .termui.color import ColorBlackOnGold, ColorBlackOnOrange, ColorGold
from .termui.control import Border
from .termui.dialog import DialogFieldCheckbox, DialogFieldLabel, DialogFieldText
from .termui.list_control import ListControl, ListEntry
from .termui.text_browser import TextBrowser
from .termui.ui import ControlCodes


def generic_confirm_template(
    method,
    template,
    *args,
    provider,
    category,
    subcategory,
    resource_type,
    summary=None,
    custom_callback=None,
    refresh=True,
    resource_identifier=None,
    undoable=False,
    can_force=False,
    extra_fields=None,
    action_name="Delete",
    success_template=None,
    from_what=None,
    from_what_name=None,
    **kwargs,
):
    """
    Creates a confirmation sequence.

    Returns a function that, when called, begins the confirmation process. The returned function requires no arguments, but takes any
    amount for legacy reasons.

    Parameters
    ----------
    method : str
        The name of the method to call for the provider when the action is confirmed.
    template : TemplateDict
        A TemplateDict which describes the keyword arguments to pass to method.
    custom_callback : Callable, optional
        A replacement callback for handling the confirmation.
    refresh : bool, default: True
        Whether to call refresh_data() after executing the action.
    category : str
        API call category, for logging purposes. Usually a pretty printed name for the AWS service that
        is being used.
    subcategory : str
        API call subcategory, for logging purposes. Usually the type of the AWS resource being managed.
    provider : str
        The name of the boto3 provider which contains the definition of the method.
    resource_type : str
        The name of the type of resource being managed.
    resource_identifier : str, optional
        Templatable via SelectionAttribute. Override for resource identifier if something other than the
        primary key of the current resource should be used.
    undoable : bool, default: False
        Cosmetic, to display a warning about data loss if not set to true.
    can_force : bool, default: False
        Adds an additional force checkbox to the confirmation dialog if set. The value of this checkbox
        is passed as "force" to the callback, and can be represented by a ForceFlag object in the template.
    summary : str, optional
        A short description of the action being performed for logging purposes. Defaults to a concatenation
        of action name and resource type (eg. 'Delete EC2 instance')
    extra_fields : dict, optional
        A map of additional fields to add to the confirmation dialog, where the key is the name of the field,
        and the value is a DialogField instance. These fields are then passed as keyword arguments to the
        callback, where the argument name is the key of the field in the dict. The value of extra fields can
        also be referred to via a FieldValue object in the template.
    action_name : str, default: "Delete"
        The name of the action being performed if the dialog is confirmed. Defaults to 'Delete' as that is both
        the original purpose of confirmation dialogs and the most common action requiring a confirmation.
    success_template : str, optional:
        A formattable string that will be displayed as a message if the operation succeeds. The format method
        will be called on the string, where index 0, and the named field 'resource' will be available as
        arguments containing the resource primary key. All keyword arguments passed to the method will also
        be passed as-is to this invocation of format. Attempts to construct a grammatically correct but not
        particularly pleasing to read message if omitted.
    from_what : str, optional
        The from_what option of a DeleteResourceDialog.
    from_what_name : str, optional
        Templatable via SelectionAttribute. The from_what_name option of a DeleteResourceDialog.

    Returns
    -------
    callable
        A function that, when called, begins the confirmation process.
    """

    def fn(self, *args):
        nonlocal category, subcategory, resource_type, provider, summary, extra_fields, success_template

        summary = summary or f"{action_name} {resource_type}"
        if extra_fields is None:
            extra_fields = {}
        success_template = (
            success_template
            or f"Performing action '{action_name.lower()} {resource_type}' on resource {'{0}'}..."
        )

        rid = ""
        if resource_identifier is not None:
            rid = resource_identifier
            if hasattr(rid, "resolve") and callable(rid.resolve):
                # Cannot insert field values at this time.
                rid = rid.resolve(self.selection, **kwargs)
        elif hasattr(self, "primary_key") and self.primary_key is not None:
            rid = self.selection[self.primary_key]

        def callback(**cb_kwargs):
            if custom_callback is not None:
                custom_callback(**cb_kwargs)
            else:
                Common.generic_api_call(
                    provider,
                    method,
                    template.resolve(self.selection, **kwargs, **cb_kwargs),
                    summary,
                    category,
                    subcategory=subcategory,
                    success_template=success_template,
                    resource=rid,
                    **cb_kwargs,
                )
            if refresh:
                self.refresh_data()

        fwn = from_what_name
        if fwn is not None and hasattr(fwn, "resolve") and callable(fwn.resolve):
            # Cannot insert field values at this time.
            fwn = fwn.resolve(self.selection, **kwargs)

        DeleteResourceDialog.opener(
            caller=self,
            resource_type=resource_type,
            resource_identifier=rid,
            callback=callback,
            can_force=can_force,
            extra_fields=extra_fields,
            action_name=action_name,
            from_what=from_what,
            from_what_name=fwn,
            **kwargs,
        )

    return fn


class RegistryHelper:
    """
    Helper class for control registry.

    Attributes
    ----------
    opener : callable
        Opener function of the control being registered.
    selector : callable
        Selector function of the control being registered.
    """

    def __init__(self, opener=None, selector=None):
        self.opener = opener
        self.selector = selector


def tag_finder_generator(tagname, default="", taglist_key="Tags"):
    """
    Generates a predicate function which can locate a tag in a raw AWS object. Useful if certain important fields are actually just tags.
    I'm looking at you, EC2 instance names.

    Parameters
    ----------
    tagname : str
        The name of the tag to find.
    default : str, default=""
        The default value of the tag, if not found.
    taglist_key : str, default="Tags"
        The name for the list of tags in the object. Believe it or not, this is less than consistent among the AWS services.

    Returns
    -------
    callable(dict-like) -> str
        A function which finds a tag in a dict-like object when called.
    """

    def fn(entry, *args):
        if taglist_key in entry:
            for tag in entry[taglist_key]:
                if tag["Key"] == tagname:
                    return tag["Value"]
        return default

    return fn


def tagged_column_generator(
    name,
    tagname,
    weight=None,
    sort_weight=None,
    hidden=False,
    size=30,
    default="",
    taglist_key="Tags",
):
    """
    Shortcut for generating a field for tag_finder_generator. Eliminates a minor amount of code duplication.

    Parameters
    ----------
    name : str
        The name of the tagged field.
    tagname : str
        The name of the tag to find.
    weight : int, optional
        The weight of the column.
    sort_weight : int, optional
        The weight of the column for sorting.
    size : int, default=30
        The size of the column in characters.
    hidden : bool, default=False
        Is this a hidden column?
    default : str
        The default value if the tag is not found.
    taglist_key : str
        The key for the tag in the JSON document.

    Returns
    -------
    dict
        A dict to insert into the columns map of a ResourceLister.
    """
    ret = {
        name: {
            "path": tag_finder_generator(
                tagname, default=default, taglist_key=taglist_key
            ),
            "hidden": hidden,
        }
    }
    if not hidden:
        if weight is not None:
            ret[name]["weight"] = weight
        ret[name]["size"] = size
    if sort_weight is not None:
        ret[name]["sort_weight"] = sort_weight
    return ret


class StopLoadingData(Exception):
    """
    Custom exception that is thrown when the data fetch thread of a ResourceLister should stop operating.
    """


class OpenableListControl(ListControl):
    """
    A ListControl which is aware of the AWSC session. Implements alternate constructor functions for use in different AWSC-specific contexts.

    Attributes
    ----------
    prefix : str
        The class-specific prefix for the color scheme entries used by this control.
    title : str
        The title of the control to display in the title bar.
    describer : callable
        Callback to call for the default describe action of the control. Describe action will not be automatically generated if this is None.
        In practice, this attribute exists with the expectation that it will point to the opener function of another OpenableListControl or
        GenericDescriber derivative. The current selection of the OpenableListControl is passed in the selection keyword argument.
    selector_cb : callable
        If set, the list control is opened in selector mode. In selector mode, the ENTER hotkey is forcibly replaced with a hotkey that calls
        an external callback passed to the initializer in the selector_cb keyword argument. This exists so we can piggyback existing resource
        listers, such as those for EC2 Instance Classes as a control for selecting from a list of valid values for the value of a dialog field,
        such as a dialog that can be used to create a new EC2 Instance with a specified EC2 Instance Class.
    hotkey_display : awsc.info.HotkeyDisplay
        The hotkey display block associated with this list control. Automatically managed by this control based on its hotkeys. Each control
        is expected to manage its own hotkey display that is then placed in the top right corner of the screen by the session manager.
    """

    prefix = "CHANGEME"
    title = "CHANGEME"
    describer: None | Callable[[], None] = None

    @classmethod
    def opener(cls, **kwargs):
        """
        Session-aware initializer for this class. Creates an OpenableListControl object with the default alignment settings for the awsc
        layout.

        Returns
        -------
        list(awsc.base_control.OpenableListControl, awsc.info.HotkeyDisplay)
            A new instance of this class, and its associated hotkey display.
        """
        instance = cls(*default_args(), weight=0, **kwargs)
        instance.border = Common.default_border(cls.prefix, cls.title, None)
        return [instance, instance.hotkey_display]

    @classmethod
    def selector(cls, callback, **kwargs):
        """
        Session-aware initializer for this class which opens it in selector mode. Otherwise, functionally identical to the opener classmethod.

        Parameters
        ----------
        callback : callable
            The selector callback.

        Returns
        -------
        list(awsc.base_control.OpenableListControl, awsc.info.HotkeyDisplay)
            A new instance of this class, and its associated hotkey display.
        """
        return cls.opener(**{"selector_cb": callback, **kwargs})

    def __init__(self, *args, selector_cb=None, **kwargs):
        kwargs["color"] = Common.color(f"{self.prefix}_generic", "generic")
        kwargs["selection_color"] = Common.color(
            f"{self.prefix}_selection", "selection"
        )
        kwargs["title_color"] = Common.color(f"{self.prefix}_heading", "column_title")
        super().__init__(*args, **kwargs)
        self.selector_cb = selector_cb
        self.hotkey_display = HotkeyDisplay.opener(caller=self)
        if selector_cb is not None:
            self.add_hotkey("KEY_ENTER", self.select_and_close, "Select and close")
        elif self.describer is not None:
            self.add_hotkey("KEY_ENTER", self.describe, "Describe")
        if self.describer is not None:
            self.add_hotkey("d", self.describe, "Describe")

    def describe(self, _):
        """
        Describe autogenerated action hotkey callback.
        """
        if self.selection is not None and self.describer is not None:
            # pylint: disable=not-callable # Yes it is.
            Common.Session.push_frame(self.describer(selection=self.selection))

    def select_and_close(self, _):
        """
        Select autogenerated action hotkey callback if control is in selector mode.
        """
        if self.selection is not None and self.selector_cb is not None:
            self.selector_cb(self.selection["name"])
            Common.Session.pop_frame()


# TODO: Reimplement resource listers as subclasses of OpenableListControl for consistency.
# TODO: Improve implementation of ResourceListerBase by moving class configuration attributes (eg. primary_key) to class attributes.
class ResourceListerBase(ListControl):
    """
    Base resource lister object. Implements basic concepts about fetching and handling AWS resources as lists.

    Attributes
    ----------
    primary_key : str
        The primary identifying field in each list entry. ListControls display the results of list operations, which will return the primary
        key of each resource in most AWS APIs. We can then use this primary key for get, update and delete operations through different
        controls.
    load_counter : int
        The number of items loaded by the control, for debug purposes.
    dialog_mode : bool
        Indicates whether a modal dialog is present over this ListControl. Having a modal dialog present will prevent automatic refresh
        operations.
    closed : bool
        Flag to indicate whether this control was closed. Used to notify the refresh thread that it should stop refreshing as there is
        no longer any point in doing so.
    next_marker : str
        An AWS next_marker, if paging from the AWS API should continue.
    next_marker_arg : str
        The argument name for the next marker in AWS API calls.
    auto_refresh_last : datetime.datetime
        Represents the last time an automatic refresh happened.
    """

    def __init__(self, *args, **kwargs):
        if not hasattr(self, "primary_key"):
            self.primary_key = None
        self.load_counter = 1
        self.dialog_mode = False
        self.closed = False
        if not hasattr(self, "next_marker"):
            self.next_marker = None
        if not hasattr(self, "next_marker_arg"):
            self.next_marker_arg = None
        self.auto_refresh_last = datetime.datetime.now()
        super().__init__(*args, **kwargs)

    def on_close(self):
        """
        Cleanup hook. Called when the control is closed.
        """
        self.closed = True

    def auto_refresh(self):
        """
        Automatic refresh function stub for implementation in subclasses.
        """

    def asynch(self, fn, *args, clear=False, **kwargs):
        """
        Async wrapper. Starts a data loading thread.

        Parameters
        ----------
        fn : callable
            A function that returns a list of new entries or a generator which yields list entries. This function will be executed in a new
            thread and each entry returned or yielded by this function will be pushed into the new entries cross-thread storage. All
            positional and keyword arguments received in the call to asynch, with the exception of clear, will be passed to this function.
        *args : list
            A list of positional arguments to pass to fn.
        clear : bool, default=False
            If True, request an asynchronous clear of existing list entries in addition to adding new ones.
        **kwargs : dict
            A map of keyword arguments to pass to fn.
        """
        try:
            thread = threading.Thread(
                target=self.async_inner,
                args=args,
                kwargs={**kwargs, "fn": fn, "clear": clear},
                daemon=True,
            )
            thread.start()
        except Exception as error:
            Common.Session.set_message(
                f"Failed to start AWS query thread: {str(error)}",
                Common.color("message_error"),
            )

    def before_paint_critical(self):
        super().before_paint_critical()
        if "thread_error" in self.thread_share:
            Common.Session.set_message(
                self.thread_share["thread_error"], Common.color("message_error")
            )
            del self.thread_share["thread_error"]

    def async_inner(self, *args, fn, clear=False, **kwargs):
        """
        Inner section of asynch. This function handles the logic of executing the data generator function.

        Parameters
        ----------
        *args : list
            A list of positional arguments to pass to fn.
        fn : callable
            A function that returns a list of new entries or a generator which yields list entries. This function will be executed in this
            thread and each entry returned or yielded by this function will be pushed into the new entries cross-thread storage. All
            positional and keyword arguments received in the call to asynch, with the exception of clear, will be passed to this function.
            The function may raise a StopLoadingData exception to gracefully abort the loading of data.
        clear : bool, default=False
            If True, request an asynchronous clear of existing list entries in addition to adding new ones.
        **kwargs : dict
            A map of keyword arguments to pass to fn.
        """
        try:
            with self.mutex:
                if clear or (
                    not hasattr(self, "primary_key") or self.primary_key is None
                ):
                    self.thread_share["clear"] = True

            for data in fn(*args, **kwargs):
                with self.mutex:
                    self.thread_share["new_entries"].extend(data)

        except StopLoadingData:
            pass
        except (
            Exception
        ) as error:  # pylint: disable=broad-except # Purpose of this function is extremely generic, needs a catch-all
            self.thread_share[
                "thread_error"
            ] = f"Refresh thread execution failed: {type(error).__name__}: {str(error)}"
            Common.error(
                f"Refresh thread execution failed: {type(error).__name__}: {str(error)}\n{traceback.format_exc()}",
                "Refresh Thread Error",
                "Core",
                set_message=False,
            )
            Common.Session.set_message(
                "Refresh thread execution failed", Common.color("message_error")
            )
        with self.mutex:
            self.thread_share["updating"] = False
            self.thread_share["finalize"] = True

    def handle_finalization_critical(self):
        """
        Finalization hook. Handles the insertion and sorting of new entries.
        """
        if hasattr(self, "primary_key") and self.primary_key is not None:
            new_entries = []
            for entry in self.entries:
                if entry[self.primary_key] in self.thread_share["acquired"]:
                    new_entries.append(entry)
            self.entries = new_entries[:]
            self.sort()

    def handle_new_entries_critical(self, entries):
        if not hasattr(self, "primary_key") or self.primary_key is None:
            super().handle_new_entries_critical(entries)
        else:
            for new in entries:
                found = False
                for old in self.entries:
                    if old[self.primary_key] == new[self.primary_key]:
                        old.mutate(new)
                        found = True
                        break
                if not found:
                    self.entries.append(new)
                self.thread_share["acquired"].append(new[self.primary_key])

    def get_data_generic(
        self,
        resource_key,
        list_method,
        list_kwargs,
        item_path,
        column_paths,
        hidden_columns,
        next_marker_name,
        next_marker_arg,
        *args,
    ):
        """
        Generic data retrieval function designed to interact with the AWS API.

        Parameters
        ----------
        resource_key : str
            A boto3 provider name.
        list_method : str
            The method to call for acquiring a list of entries within the boto3 provider.
        list_kwargs : dict
            A map of keyword arguments to pass to the method call.
        item_path : str
            Jq expression which, when applied to the json response from the AWS API, results in a json list of entries.
        column_paths : dict
            A mapping of column names to jq expressions. Each expression will be run in the context of a single entry in the list acquired
            by running the item_path jq expression.
        hidden_columns : dict
            Same as column_paths, except these columns are hidden from display.
        next_marker_name : str
            The name of the next marker field in the AWS API response.
        next_marker_arg : str
            The keyword argument for the next marker field for the list method.

        Yields
        ------
        list(ListEntry)
            A list of new entries. The generator yields after each API call the result of that single API call.
        """
        try:
            provider = Common.Session.service_provider(resource_key)
        except KeyError:
            Common.error(
                "boto3 does not recognize provider",
                "Invalid provider",
                "Core",
                subcategory="ResourceListerBase",
                api_provider=resource_key,
                classname=type(self).__name__,
            )
            return []
        if callable(list_kwargs):
            list_kwargs = list_kwargs()
        ret = []
        next_marker = None
        while next_marker != "":
            if self.closed:
                raise StopLoadingData
            it_list_kwargs = list_kwargs.copy()
            if next_marker is not None and next_marker_arg is not None:
                it_list_kwargs[next_marker_arg] = next_marker
            try:
                response = getattr(provider, list_method)(**it_list_kwargs)
            except botoerror.ClientError as error:
                Common.clienterror(
                    error,
                    "List Resources",
                    "Core",
                    subcategory="ResourceListerBase",
                    api_provider=resource_key,
                    api_method=list_method,
                    api_kwargs=it_list_kwargs,
                )
                # pylint: disable=raise-missing-from # StopLoadingData is a special exception to stop this generator from being used.
                raise StopLoadingData

            items = (
                Common.Session.jq(item_path)
                .input(text=json.dumps(response, default=datetime_hack))
                .first()
            )
            if items is None:
                Common.error(
                    "get_data_generic returned None",
                    "Get Data Generic returned None",
                    "Core",
                    set_message=False,
                    api_provider=resource_key,
                    api_method=list_method,
                    api_kwargs=list_kwargs,
                    item_path=item_path,
                    api_response=response,
                )
                return []

            for item in items:
                if self.closed:
                    raise StopLoadingData
                init = {}
                for column, path in {**column_paths, **hidden_columns}.items():
                    if callable(path):
                        init[column] = path(item)
                    else:
                        try:
                            init[column] = Common.Session.jq(path).input(item).first()
                        except StopIteration:
                            init[column] = ""
                list_entry = ListEntry(**init)
                list_entry.controller_data = item
                if self.matches(list_entry, *args):
                    ret.append(list_entry)
            if self.closed:
                raise StopLoadingData
            yield ret
            if self.closed:
                raise StopLoadingData
            self.load_counter += 1
            if next_marker_name is not None and next_marker_name in response:
                next_marker = response[next_marker_name]
                ret = []
            else:
                next_marker = ""

    def matches(self, list_entry, *args):
        """
        For pre-filtering resource listers, this function decides whether a list entry being generated matches the criteria for being
        displayed on the control. Used by relational lister subclasses to check for relations.

        Returns
        -------
        bool
            Whether an entry matches the criteria for insertion.
        """
        return True

    def empty(self, _):
        """
        Callback used to generate an empty string. May be unused now.

        Returns
        -------
        str
            An empty string.
        """
        return ""


class DialogFieldResourceListSelector(DialogFieldText):
    """
    A dialog field which allows selecting a single resource in a ResourceLister as the value of the field.

    Attributes
    ----------
    left : int
        How far to the right is the textfield scrolled.
    text : str
        The text in the field. The primary key of the resource will be used for this purpose.
    label : str
        The label of the field.
    color : awsc.termui.color.Color
        The color of the field when not selected.
    label_color : awsc.termui.color.Color
        The color of the field's label.
    selected_color : awsc.termui.color.Color
        The color of the field when it is selected.
    selector_class : object
        An object with a selector method, which takes a callable as an argument and returns a frame list for the session.
        This object will be opened when the user interacts with the field.
    primary_key : str
        If set, overrides the primary key of the selector class.
    """

    def __init__(
        self,
        selector_class,
        label,
        default="",
        color=ColorBlackOnOrange,
        selected_color=ColorBlackOnGold,
        label_color=ColorGold,
        label_min=0,
        primary_key=None,
    ):
        super().__init__(
            label,
            text=default,
            color=color,
            selected_color=selected_color,
            label_color=label_color,
            label_min=label_min,
            password=False,
            accepted_inputs=None,
        )
        self.selector_class = selector_class
        self.primary_key = primary_key

    def selector_callback(self, entry):
        """
        The callback the lister will call when an entry is selected.

        Parameters
        ----------
        entry : str
            The entry that was selected in the lister.
        """
        self.text = entry

    def input(self, inkey):
        if inkey.is_sequence:
            if inkey.name == "KEY_BACKSPACE":
                inkey.name = "KEY_DELETE"
            if inkey.name in (
                "KEY_LEFT",
                "KEY_RIGHT",
                "KEY_HOME",
                "KEY_END",
                "KEY_DELETE",
            ):
                return super().input(inkey)
            if inkey.name == "KEY_ENTER":
                kwa = {}
                if self.primary_key is not None:
                    kwa["primary_key"] = self.primary_key
                Common.Session.push_frame(
                    self.selector_class.selector(self.selector_callback, **kwa)
                )
                Common.Session.ui.dirty = True
                return True
        return False

    def paint(self, x0, x1, y, selected=False):
        x = x0
        Common.Session.ui.print(self.label, xy=(x, y), color=self.label_color)
        x += max(len(self.label) + 1, self.label_min)
        space = x1 - x + 1
        text = self.text + " ↲"
        if self.left >= len(text):
            self.left = 0
        text = text[
            self.left : (
                self.left + space if len(text) > self.left + space else len(text)
            )
        ]
        Common.Session.ui.print(
            text, xy=(x, y), color=self.selected_color if selected else self.color
        )


# TODO: Create ResourceLister.naming_template for reducing method spam.
class SingleNameDialog(SessionAwareDialog):
    """
    A convenience class for quickly creating a dialog for entering a single string in a textfield.

    Attributes
    ----------
    what : str, default="name"
        The piece of data being entered into the field.
    error_label : awsc.termui.dialog.DialogFieldLabel
        An error label, where validation errors may be displayed.
    name_field : awsc.termui.dialog.DialogFieldText
        The textfield where the name or other data may be typed.
    caller : awsc.termui.control.Control
        The control which opened this dialog.
    callback : callable
        Callback to call when the dialog is confirmed.
    """

    def __init__(
        self,
        parent,
        title,
        callback,
        *args,
        label="Name:",
        what="name",
        subject="",
        default="",
        caller=None,
        accepted_inputs=None,
        **kwargs,
    ):
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            title,
            Common.color("modal_dialog_border_title"),
        )
        super().__init__(
            parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=caller,
            *args,
            **kwargs,
        )
        self.what = what
        for_part = f" for {subject}" if subject != "" else ""
        self.set_title_label(f"Enter {what}{for_part}")
        self.name_field = DialogFieldText(
            label,
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
            accepted_inputs=accepted_inputs,
        )
        self.add_field(self.name_field)
        self.caller = caller
        self.callback = callback

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = f"You must enter a {self.what}."
            return
        self.callback(self.name_field.text)
        super().accept_and_close()

    def close(self):
        if self.caller is not None:
            self.caller.refresh_data()
        super().close()


class SingleSelectorDialog(SessionAwareDialog):
    """
    A convenience class for quickly creating a dialog for selecting a single resource.

    Attributes
    ----------
    error_label : awsc.termui.dialog.DialogFieldLabel
        An error label, where validation errors may be displayed.
    resource_selector : awsc.base_control.DialogFieldResourceListSelector
        The field that allows selecting a single resource.
    caller : awsc.termui.control.Control
        The control which opened this dialog.
    callback : callable
        Callback to call when the dialog is confirmed.
    """

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        title,
        resource_type,
        action_name,
        selector_class,
        callback,
        *args,
        caller=None,
        **kwargs,
    ):
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            title,
            Common.color("modal_dialog_border_title"),
        )
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            caller=caller,
            **kwargs,
        )
        self.set_title_label(f"Select {resource_type} for {action_name} action")

        if isinstance(selector_class, str):
            # Has a .selector attribute, so it should be fine.
            selector_class = Common.Session.control_registry[selector_class]

        self.resource_selector = DialogFieldResourceListSelector(
            selector_class,
            f"{resource_type}: ",
            "",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.resource_selector)
        self.caller = caller
        self.callback = callback

    def accept_and_close(self):
        if self.resource_selector.text == "":
            self.error_label.text = "You must select a resource."
            return
        self.callback(self.resource_selector.text)
        super().accept_and_close()

    def close(self):
        if self.caller is not None:
            self.caller.refresh_data()
        super().close()


class ForceFlag:
    """
    Represents the value of the force checkbox of a DeleteResourceDialog.
    """

    def resolve(self, selection, **kwargs):
        """
        Resolves the ForceFlag into a value.

        Parameters
        ----------
        selection : termui.list_control.ListEntry
            ForceFlag ignores the selection argument.
        **kwargs: dict, optional
            If this contains an argument named 'force', its value will be used for resolution.

        Returns
        -------
        bool
            Returns the value received in the 'force' keyword argument, or False if it is not present.
        """
        return False if "force" not in kwargs else kwargs["force"]


class FieldValue:
    """
    Represents the value of a field passed as an extra field to a dialog template.

    Attributes
    ----------
    field_name : str
        The name of the field whose value will be used when resolving.
    """

    def __init__(self, field_name):
        """
        Initializes a FieldValue.

        Parameters
        ----------
        field_name : str
            The name of the field whose value will be used when resolving.
        """
        self.field_name = field_name

    def resolve(self, selection, **kwargs):
        """
        Resolves the FieldValue into a value.

        Parameters
        ----------
        selection : termui.list_control.ListEntry
            FieldValue ignores the selection argument.
        **kwargs: dict, optional
            If this contains the field specified in the field_name attribute, its value will be used for resolution.

        Returns
        -------
        any
            The type of the returned value matches the value type of the DialogField specified in field_name. If the field
            is not present in the keyword arguments passed to this function, an empty (None) value is returned.
        """
        return None if self.field_name not in kwargs else kwargs[self.field_name].value


class SelectorTemplateField(FieldValue):
    """
    Convenience class for handling the value of the selection template field.
    """

    def __init__(self):
        super().__init__("value")


class SelectionAttribute:
    """
    Represents a templatable attribute from a ListEntry row.

    Attributes
    ----------
    attribute : str
        The name of the column (attribute) to use from the ListEntry for templating.
    """

    def __init__(self, attribute):
        """
        Initializes a SelectionAttribute.

        Parameters
        ----------
        attribute : str
            The name of the column (attribute) to use from the ListEntry for templating.
        """
        self.attribute = attribute

    def resolve(self, selection, **kwargs):
        """
        Resolves the SelectionAttribute into a value.

        Parameters
        ----------
        selection : termui.list_control.ListEntry
            A ListEntry to use as the row of values for resolving the SelectionAttribute.
        **kwargs: dict, optional
            All other keyword arguments are ignored.

        Returns
        -------
        any
            The type of the returned value matches the type of value stored in the ListEntry. This is ideally,
            but not guaranteed to be a string.
        """
        return selection[self.attribute]


class SelectionControllerDataAttribute:
    """
    Represents a templatable attribute from a ListEntry row's controller data.

    Attributes
    ----------
    attribute : str
        The name of the column (attribute) to use from the ListEntry for templating.
    """

    def __init__(self, attribute):
        """
        Initializes a SelectionControllerDataAttribute.

        Parameters
        ----------
        attribute : str
            The name of the column (attribute) to use from the ListEntry's controller data for templating.
        """
        self.attribute = attribute

    def resolve(self, selection, **kwargs):
        """
        Resolves the SelectionControllerDataAttribute into a value.

        Parameters
        ----------
        selection : termui.list_control.ListEntry
            A ListEntry to use as the row of values for resolving the SelectionControllerDataAttribute.
        **kwargs: dict, optional
            All other keyword arguments are ignored.

        Returns
        -------
        any
            The type of the returned value matches the type of value stored in the ListEntry's controller data.
        """
        return selection.controller_data[self.attribute]


class TemplateList(list):
    """
    A list which can be templated from a ResourceLister control.
    """

    def __init__(self, source=None):
        """
        Initializes a TemplateList.

        Parameters
        ----------
        source : list, optional
            If provided, initializes the TemplateList by recursively converting the provided list.
        """
        super().__init__()
        if source is not None:
            for item in source:
                if isinstance(item, dict):
                    self.append(TemplateDict(item))
                elif isinstance(item, list):
                    self.append(TemplateList(item))
                else:
                    self.append(item)

    def resolve(self, selection, **kwargs):
        """
        Resolves the TemplateList into a list.

        Parameters
        ----------
        selection : termui.list_control.ListEntry
            The selection of a ListControl object to use for resolving SelectionAttribute objects.
        **kwargs : dict, optional
            A list of additional arguments for ForceFlag and FieldValue objects, as passed to the kwargs of
            a DeleteResourceDialog callback.

        Returns
        -------
        list
            A list with the templatable elements recursively resolved to values.
        """
        ret = []
        for item in self:
            if hasattr(item, "resolve"):
                ret.append(item.resolve(selection, **kwargs))
            else:
                ret.append(item)
        return ret


class TemplateDict(dict):
    """
    A dict which can be templated from a ResourceLister control.
    """

    def __init__(self, source=None):
        """
        Initializes a TemplateDict.

        Parameters
        ----------
        source : dict, optional
            If provided, initializes the TemplateDict by recursively converting the provided dict.
        """
        super().__init__()
        if source is not None:
            for key, value in source.items():
                if isinstance(value, dict):
                    self[key] = TemplateDict(value)
                elif isinstance(value, list):
                    self[key] = TemplateList(value)
                else:
                    self[key] = value

    def resolve(self, selection, **kwargs):
        """
        Resolves the TemplateDict into a dict.

        Parameters
        ----------
        selection : termui.list_control.ListEntry
            The selection of a ListControl object to use for resolving SelectionAttribute objects.
        **kwargs : dict, optional
            A list of additional arguments for ForceFlag and FieldValue objects, as passed to the kwargs of
            a DeleteResourceDialog callback.

        Returns
        -------
        dict
            A dict with the templatable elements recursively resolved to values.
        """
        ret = {}
        for key, value in self.items():
            if hasattr(value, "resolve"):
                ret[key] = value.resolve(selection, **kwargs)
            else:
                ret[key] = value
        return ret


class ResourceRefByCommand:
    """
    A ResourceRefByCommand object refers to any openable control which is registered in the Session commander_options by the command.

    Calling an instance of this object is equivalent to calling the opener classmethod of that class.
    """

    def __init__(self, cmd):
        self.cmd = cmd

    def __call__(self, *args, **kwargs):
        return Common.Session.commander_options[self.cmd](*args, **kwargs)


class ResourceRefByClass:
    """
    A ResourceRefByClass object refers to any openable control which is registered in the Session control_registry by class name.

    Calling an instance of this object is equivalent to calling the opener classmethod of that class.
    """

    def __init__(self, cmd):
        self.cmd = cmd

    def __call__(self, *args, **kwargs):
        return Common.Session.control_registry[self.cmd].opener(*args, **kwargs)


class ResourceLister(ResourceListerBase):
    """
    The meaty ResourceLister is the object that handles all of the resource listing from AWS. Most basic resource listers will be direct
    subclasses with minor configuration.

    Attributes
    ----------
    command_palette : list(str)
        A list of commands for opening this particular resource lister. This may be any proper name for the AWS resource, such as `queue`,
        a shorthand, such as `sqs`, or any commonly known alternative name.
    resource_type : str
        The human-readable name of the resource type.
    main_provider : str
        The provider the list_method uses in boto3.
    category : str
        A human-readable name of the category of AWS resources this lister belongs to, for logging purposes.
    subcategory : str
        A human-readable name of the subcategory of AWS resources this lister belongs to, for logging purposes.
    list_method : str
        The method to call to list the resources.
    list_kwargs : dict
        The kwargs to pass to the list method. May be modified in the initializer for dynamic parameters.
    item_path : str
        The path where an item can be retrieved.
    columns : dict
        A map of column specification. Each column's key is the name of the column, while the value is a dict with the following fields:
        - path (required): The jq path for the column within the item.
        - size (required if not hidden): The size of the column.
        - weight (int, default 0): The weight of the column for sorting the columns in order.
        - hidden (bool, default false): Whether the column is hidden.
        - sort_weight (int, optional): If present, the weight of the column for sorting rows. Columns without a sort weight are not used
          for sorting.
    primary_key : str
        The field which contains the unique identifier of the resource. For most resources, this is the same field every time. Not always,
        unfortunately.
    describe_command : callable or str
        The command to run for the describe autogenerated action. If str, this is a reference to another, existing hotkey. If callable, the
        function is called and the return value is pushed as a new frame (so ideally, use an opener classmethod as the thing to pass to this).
        The describe command will be assigned the d hotkey, as well as the ENTER hotkey is no open command is present.
    describe_kwargs : dict
        Additional keyword arguments to pass to the describe command.
    describe_selection_arg : str, default="entry"
        The name of the keyword argument where the selection of this list control should be passed on describe.
    open_command : callable or str
        The command to run for the open autogenerated action. If str, this is a reference to another, existing hotkey. If callable, the
        function is called and the return value is pushed as a new frame (so ideally, use an opener classmethod as the thing to pass to this).
        The open command will be assigned the ENTER hotkey. If not present, the describe_command is instead assigned to the ENTER key.
    open_kwargs : dict
        Additional keyword arguments to pass to the open command.
    open_selection_arg : str, default="entry"
        The name of the keyword argument where the selection of this list control should be passed on open.
    """

    prefix = "CHANGEME"
    title = "CHANGEME"
    command_palette: List[str] = []

    resource_type = "CHANGEME"
    main_provider = "CHANGEME"
    category = "CHANGEME"
    subcategory = "CHANGEME"
    list_method = "CHANGEME"
    list_kwargs: Dict[str, Any] = {}
    item_path = ".ChangeMe"
    columns: Dict[str, Dict[str, Any]] = {}
    primary_key: str | None = "name"

    describe_selection_arg = "entry"
    describe_command: None | Callable | str = None
    describe_kwargs: Dict[str, Any] = {}
    open_selection_arg = "entry"
    open_command: None | Callable | str = None
    open_kwargs: Dict[str, Any] = {}

    additional_commands: Dict[str, Dict[str, Any]] = {}

    class Autocommand:
        """
        Class decorator that automatically assigns a hotkey to a function upon instantiation of the target class.

        This is fuelled by the worst python can offer, I wish you the best of luck in understanding its source.

        Usage: @ResourceLister.Autocommand(command_class, key, tooltip=None, selection_arg=\"entry\", **kwargs)

        Attributes
        ----------
        command_class : str
            The name of the class where the decorated class becomes a command.
        key : str
            The key to press to trigger the command. See HotkeyControl.add_hotkey for more details.
        tooltip : str, optional
            The tooltip to associate with the command. See HotkeyControl.add_hotkey for more details.
        selection_arg : str, default="entry"
            The argument name for passing the selection to the decorated class initializer.
        **kwargs: dict
            A dict of additional keyword arguments to pass to the decorated class initializer.
        """

        _autocommands: Dict[str, Dict[str, Dict[str, Any]]] = {}

        def __init__(
            self, command_class, key, tooltip=None, selection_arg="entry", **kwargs
        ):
            self.command_class = command_class
            self.key = key
            self.tooltip = tooltip
            self.selection_arg = selection_arg
            self.kwargs = kwargs

        def __call__(self, cls):
            if self.command_class not in self.__class__._autocommands:
                self.__class__._autocommands[self.command_class] = {}
            self.__class__._autocommands[self.command_class][self.key] = {
                "command": cls.opener,
                "tooltip": self.tooltip,
                "selection_arg": self.selection_arg,
                "kwargs": self.kwargs,
            }
            return cls

        @classmethod
        def auto_command(cls, self):
            """
            Performs automatically hotkeying up a ResourceLister subclass.

            Inserts all inherited commands automatically. Insertion is done in reverse of MRO - therefore, higher classes on the inheritance tree,
            such as ResourceLister, insert their commands first. This allows subclasses to overwrite these commands.

            Parameters
            ----------
            self : awsc.termui.base_control.ResourceLister
                The class being initialized.
            """
            import inspect

            for classobj in reversed(inspect.getmro(self.__class__)):
                classname = classobj.__name__
                if classname in cls._autocommands:
                    for hotkey, definition in cls._autocommands[classname].items():
                        self.additional_commands[hotkey] = definition

    def autohotkey_condition(self, hotkey):
        return self.selector_cb is None

    def confirm_template(self, *args, hotkey=None, hotkey_tooltip=None, **kwargs):
        """
        Creates a confirmation sequence.

        Can be used as a shorthand for registering a confirmation dialog as a hotkey. Returns a function that,
        when called, begins the confirmation process. The returned function requires no arguments, but takes any
        amount for legacy reasons.

        Args and kwargs correspond to the parameters of generic_confirm_template.

        Parameters
        ----------
        hotkey : str, optional
            The hotkey for this action. This string should be a valid key for the add_hotkey function. If not
            provided, no hotkey will be added for this confirmation dialog.
        hotkey_tooltip : str, optional
            If set, passes a tooltip for the hotkey to be displayed on the hotkey display list. If omitted, the
            hotkey will be hidden from the hotkey display list.

        Returns
        -------
        callable
            A function that, when called, begins the confirmation process.
        """

        kwargs["category"] = (
            "category" in kwargs and kwargs["category"]
        ) or self.category
        kwargs["subcategory"] = (
            "subcategory" in kwargs and kwargs["subcategory"]
        ) or self.subcategory
        kwargs["resource_type"] = (
            "resource_type" in kwargs and kwargs["resource_type"]
        ) or self.resource_type
        kwargs["provider"] = (
            "provider" in kwargs and kwargs["provider"]
        ) or self.main_provider

        generic = generic_confirm_template(*args, **kwargs)

        def fn(*args):
            return generic(self, *args)

        if hotkey is not None:
            self.add_hotkey(hotkey, fn, hotkey_tooltip, True)
        return fn

    def selector_template(
        self,
        method,
        template,
        title,
        action_name,
        selector_class,
        resource_type=None,
        category=None,
        subcategory=None,
        provider=None,
        summary=None,
        success_template=None,
        caller=None,
        hotkey=None,
        hotkey_tooltip=None,
        hotkey_validated=True,
        refresh=True,
        **kwargs,
    ):
        """
        Templates a SingleSelectionDialog flow, which calls an AWS API method with the provided template at the end of the flow.

        Parameters
        ----------
        method : str
            The name of the method to call for the provider when the action is confirmed.
        template : TemplateDict
            A TemplateDict which describes the keyword arguments to pass to method.
        title : str
            The title for the selection dialog.
        action_name : str
            The action being performed.
        resource_type : str, optional
            The name of the type of resource being managed. Defaults to the resource_type set on the class.
        category : str, optional
            API call category, for logging purposes. Usually a pretty printed name for the AWS service that
            is being used. Defaults to the category set on the class.
        subcategory : str, optional
            API call subcategory, for logging purposes. Usually the type of the AWS resource being managed.
            Defaults to the subcategory set on the class.
        provider : str, optional
            The name of the boto3 provider which contains the definition of the method. Defaults to the
            main_provider set on the class.
        summary : str, optional
            A short description of the action being performed for logging purposes. Defaults to a concatenation
            of action name and resource type (eg. 'Delete EC2 instance')
        success_template : str, optional:
            A formattable string that will be displayed as a message if the operation succeeds. The format method
            will be called on the string, where index 0, and the named field 'resource' will be available as
            arguments containing the resource primary key. All keyword arguments passed to the method will also
            be passed as-is to this invocation of format. Attempts to construct a grammatically correct but not
            particularly pleasing to read message if omitted.
        hotkey : str, optional
            The hotkey for this action. This string should be a valid key for the add_hotkey function. If not
            provided, no hotkey will be added for this confirmation dialog.
        hotkey_tooltip : str, optional
            If set, passes a tooltip for the hotkey to be displayed on the hotkey display list. If omitted, the
            hotkey will be hidden from the hotkey display list.
        hotkey_validated : bool, default=True
            Whether the hotkey to add needs validation.
        refresh : bool, default: True
            Whether to call refresh_data() after executing the action.
        """

        def fn(*args):
            nonlocal category, subcategory, resource_type, provider, summary, success_template

            category = category or self.category
            subcategory = subcategory or self.subcategory
            resource_type = resource_type or self.resource_type
            provider = provider or self.main_provider
            summary = summary or f"{action_name} {resource_type}"
            success_template = (
                success_template
                or f"Performing action '{action_name.lower()} {resource_type}' on resource {'{0}'}..."
            )

            def callback(value):
                cb_kwargs = {"value": value}
                Common.generic_api_call(
                    provider,
                    method,
                    template.resolve(self.selection, **kwargs, **cb_kwargs),
                    summary,
                    category,
                    subcategory=subcategory,
                    success_template=success_template,
                    resource=self.selection[self.primary_key],
                    **cb_kwargs,
                )
                if refresh:
                    self.refresh_data()

            SingleSelectorDialog.opener(
                title,
                resource_type,
                action_name,
                selector_class,
                callback,
                *args,
                caller=self,
                **kwargs,
            )

        if hotkey is not None:
            self.add_hotkey(hotkey, fn, hotkey_tooltip, hotkey_validated)
        return fn

    @classmethod
    def register(cls):
        """
        Classmethod to register this class and all of its subclasses with the commander control.

        Also stores the class in the session control registry for other uses.

        This is why we can reach things via the command palette.
        """
        Common.Session.control_registry[cls.__name__] = RegistryHelper(
            opener=cls.opener, selector=cls.selector
        )
        for cmd in cls.command_palette:
            Common.Session.commander_options[cmd] = cls.opener
        for subcls in cls.__subclasses__():
            subcls.register()

    @classmethod
    def opener(cls, **kwargs):
        """
        Session-aware initializer for this class. Creates an ResourceLister object with the default alignment settings for the awsc
        layout.

        Returns
        -------
        list(awsc.base_control.OpenableListControl, awsc.info.HotkeyDisplay)
            A new instance of this class, and its associated hotkey display.
        """
        instance = cls(*default_args(), **default_kwargs(cls.prefix), **kwargs)
        instance.border = Common.default_border(
            cls.prefix, cls.title, instance.title_info()
        )
        return [instance, instance.hotkey_display]

    @classmethod
    def selector(cls, callback, **kwargs):
        """
        Session-aware initializer for this class for opening it in selector mode. Otherwise functionally identical to opener().

        Parameters
        ----------
        callback : callable
            The selector callback.

        Returns
        -------
        list(awsc.base_control.OpenableListControl, awsc.info.HotkeyDisplay)
            A new instance of this class, and its associated hotkey display.
        """
        return cls.opener(**{"selector_cb": callback, **kwargs})

    def title_info(self):
        """
        Title info callback for this class. The return value of this shown in the title info dialog.

        Returns
        -------
        str
            The additional info to display in the title of the control.
        """
        return None

    def __init__(self, *args, selector_cb=None, **kwargs):
        super().__init__(*args, **kwargs)
        # resource_key is now main_provider
        ResourceLister.Autocommand.auto_command(self)
        if "list_kwargs" in kwargs:
            self.list_kwargs = kwargs["list_kwargs"]
        if isinstance(self.open_command, str) and (
            not hasattr(self, "additional_commands")
            or self.open_command not in self.additional_commands
        ):
            raise AttributeError(
                "open_command refers to a key that is not in additional_commands"
            )

        if isinstance(self.describe_command, str) and (
            not hasattr(self, "additional_commands")
            or self.describe_command not in self.additional_commands
        ):
            raise AttributeError(
                "describe_command refers to a key that is not in additional_commands"
            )

        self.hidden_columns = {}
        self.column_paths = {}
        self.column_titles = {}
        self.sort_order = []
        self.column_order = []

        for name, spec in self.columns.items():
            if "hidden" in spec and spec["hidden"]:
                self.hidden_columns[name] = spec["path"]
            else:
                self.column_paths[name] = spec["path"]
                self.column_titles[name] = spec["size"]
                self.column_order.append(name)
            if "sort_weight" in spec:
                self.sort_order.append(name)
        self.column_order.sort(
            key=lambda x: self.columns[x]["weight"]
            if "weight" in self.columns[x]
            else 0
        )
        self.sort_order.sort(key=lambda x: self.columns[x]["sort_weight"])

        if "name" not in self.column_paths and "name" not in self.hidden_columns:
            raise AttributeError(
                "name entry is required in column_paths or hidden_columns"
            )

        if "primary_key" in kwargs:
            self.primary_key = kwargs["primary_key"]
        elif not hasattr(self, "primary_key") or self.primary_key is None:
            self.primary_key = "name"

        self.selector_cb = selector_cb

        if self.open_command is not None:
            open_tooltip = "Open"
            if isinstance(self.open_command, str):
                cmd = self.additional_commands[self.open_command]
                self.open_command = cmd["command"]
                self.open_selection_arg = cmd["selection_arg"]
                open_tooltip = cmd["tooltip"]
            if selector_cb is None:
                self.add_hotkey("KEY_ENTER", self.open, open_tooltip)
            else:
                self.add_hotkey("o", self.open, open_tooltip)
        if self.describe_command is not None:
            describe_tooltip = "Describe"
            if isinstance(self.describe_command, str):
                cmd = self.additional_commands[self.describe_command]
                self.describe_command = cmd["command"]
                self.describe_selection_arg = cmd["selection_arg"]
                describe_tooltip = cmd["tooltip"]
            self.add_hotkey("d", self.describe, "Describe")
            if self.open_command is None and selector_cb is None:
                self.add_hotkey("KEY_ENTER", self.describe, describe_tooltip)
        if selector_cb is not None:
            self.add_hotkey("KEY_ENTER", self.select_and_close, "Select")
        if hasattr(self, "additional_commands"):
            for key, command_spec in self.additional_commands.items():
                if "kwargs" not in command_spec:
                    command_spec["kwargs"] = {}
                self.add_hotkey(
                    key,
                    self.command_wrapper(
                        command_spec["command"],
                        command_spec["selection_arg"],
                        **command_spec["kwargs"],
                    ),
                    command_spec["tooltip"],
                )
        self.add_hotkey(ControlCodes.R, self.refresh_data, "Refresh")
        if "arn" in self.column_paths or "arn" in self.hidden_columns:
            self.add_hotkey("r", self.copy_arn, "Copy ARN")
        self.hotkey_display = HotkeyDisplay.opener(caller=self)

        self.refresh_data()

    def copy_arn(self, *args):
        """
        Copies the ARN of the selection to the clipboard if available. The ARN is the value of the arn field in the entry.
        Does nothing if the arn field is not present.
        """
        if self.selection is not None:
            pyperclip.copy(self.selection["arn"])
            Common.Session.set_message(
                "Copied resource ARN to clipboard", Common.color("message_success")
            )

    def select_and_close(self, *args):
        """
        Hotkey callback for KEY_ENTER in selector mode. Calls the selector callback and closes the control.
        """
        if (
            self.selection is not None
            and self.selector_cb is not None
            and self.primary_key is not None
        ):
            self.selector_cb(self.selection[self.primary_key])
            Common.Session.pop_frame()

    def command(self, cmd, **kwargs):
        """
        Inner section of a command hotkey callback.
        Commands are hotkeys that open new controls, and therefore require an interaction with the session.

        Commands are automatically loaded from the additional_commands dict.

        Parameters
        ----------
        cmd : callable -> list(awsc.termui.block.Block)
            Callable which returns a display stack frame.
        **kwargs : dict
            A map of additional keyword arguments to pass to cmd.
        """
        if self.selection is not None:
            frame = cmd(**kwargs)
            if frame is not None:
                Common.Session.push_frame(frame)

    def command_wrapper(self, cmd, selection_arg, **kwargs):
        """
        Wraps a command as a hotkey callback.
        Commands are hotkeys that open new controls, and therefore require an interaction with the session.

        Commands are automatically loaded from the additional_commands dict.

        Parameters
        ----------
        cmd : callable -> list(awsc.termui.block.Block)
            Callable which returns a display stack frame.
        selection_arg : str
            The name of the keyword argument where the current selection should be passed to the command.
        **kwargs : dict
            A map of additional keyword arguments to pass to cmd.
        """

        def fn(*args):
            kwargs[selection_arg] = self.selection
            self.command(
                cmd,
                pushed=True,
                caller=self,
                **kwargs,
            )

        return fn

    def describe(self, *args):
        """
        Describe is a special command automatically assigned to 'd' if describe_command is set.

        It is also assigned to ENTER if open_command is not set.
        """
        if self.describe_command is not None:
            self.command_wrapper(
                self.describe_command,
                self.describe_selection_arg,
                **self.describe_kwargs,
            )()

    def open(self, *args):
        """
        Describe is a special command automatically assigned to ENTER if open_command is set.
        """
        if self.open_command is not None:
            self.command_wrapper(self.open_command, self.open_selection_arg)()

    def get_data(self, *args, **kwargs):
        """
        Data streaming function. Yields the return value of get_data_generic one entry at a time.

        Yields
        ------
        awsc.termui.list_control.ListEntry
            A single new entry to add to the list.
        """
        for y in self.get_data_generic(
            self.main_provider,
            self.list_method,
            self.list_kwargs,
            self.item_path,
            self.column_paths,
            self.hidden_columns,
            self.next_marker,
            self.next_marker_arg,
        ):
            yield y

    def auto_refresh(self):
        if self.dialog_mode or self.primary_key is None:
            return
        if datetime.datetime.now() - self.auto_refresh_last > datetime.timedelta(
            seconds=10
        ):
            self.refresh_data()

    def refresh_data(self, *args, **kwargs):
        """
        Performs a full refresh asynchronously, fetching all data again and updating all entries as required.
        """
        with self.mutex:
            if "updating" in self.thread_share and self.thread_share["updating"]:
                return
            self.thread_share["updating"] = True
            self.thread_share["acquired"] = []
        self.auto_refresh_last = datetime.datetime.now()
        self.asynch(self.get_data)

    def sort(self):
        for sort_field in reversed(self.sort_order):
            self.entries.sort(key=attrgetter(sort_field))
        self._cache = None


class NoResults(Exception):
    """
    NoResults is an exception raised by data fetch functions to note that there were no results.
    """


class SingleRelationLister(ResourceListerBase):
    """
    SingleRelationLister parses and lists resources from a single resource's description - such as an EC2 instance. It can be used to display
    a list of resources related to a single resource.

    resource_type : str
        The human-readable name of the resource type.
    main_provider : str
        The provider the describe_method uses in boto3.
    category : str
        A human-readable name of the category of AWS resources this lister belongs to, for logging purposes.
    subcategory : str
        A human-readable name of the subcategory of AWS resources this lister belongs to, for logging purposes.
    describe_method : str
        The method to call to retrieve the single relational resource.
    describe_kwargs : TemplateDict
        The kwargs to pass to the describe method. This is essentially never expected to not depend on an input, so we define it as a
        TemplateDict. At the end of the day, whatever this is is expected to have a resolve() method that takes one positional and two
        keyword arguments and returns a dict. Selection argument received by the __init__ method is passed as the selection, as well as
        all kwargs received as the kwargs.
    object_path : str
        The path where the resource being related is found.
    resource_descriptors : list
        A list of resource specifications. Each resource specification is a dict with the following fields:
        - base_path: The jq path for where the primary keys of this related resource can be found within the resource description.
          The jq expression for base_path must always yield a list of strings.
        - type: The human readable name of the resource type.
        - describer (optional): A callable, usually a ResourceRef or opener function, for the describer of the resource type. Describe
          actions will not be allowed on resources that don't have a describer entry.
    parent_selection : awsc.termui.list_control.ListEntry
        The selection from whatever is opening this. Any dict-like object will do, ListEntry is preferred due to its getattr override.

    """

    prefix = "CHANGEME"
    title = "CHANGEME"

    resource_type = "CHANGEME"
    main_provider = "CHANGEME"
    category = "CHANGEME"
    subcategory = "CHANGEME"
    describe_method = "CHANGEME"
    describe_kwargs: TemplateDict = TemplateDict()
    object_path = ".ChangeMe"
    resource_descriptors: List[Dict[str, Any]] = []

    def title_info(self):
        """
        Title info functions are used to determine the title info after a lister has been instantiated.

        Returns
        -------
        str
            The information in the title.
        """
        return None

    @classmethod
    def opener(cls, **kwargs):
        """
        Session-aware initializer for this class. Creates an SingleRelationLister object with the default alignment settings for the awsc
        layout.

        Returns
        -------
        list(awsc.base_control.SingleRelationLister, awsc.info.HotkeyDisplay)
            A new instance of this class, and its associated hotkey display.
        """
        instance = cls(*default_args(), **default_kwargs(cls.prefix), **kwargs)
        instance.border = Common.default_border(
            cls.prefix, cls.title, instance.title_info()
        )
        return [instance, instance.hotkey_display]

    def __init__(self, parent, alignment, dimensions, *args, selection, **kwargs):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.column_order = []
        self.column_titles = {}
        self.parent_selection = selection
        self.describe_kwargs = self.describe_kwargs.resolve(selection, **kwargs)
        self.add_column("type", 30, 0)
        self.add_column("id", 120, 1)
        self.descriptor = None
        self.descriptor_raw = None
        self.add_hotkey("d", self.describe, "Describe", True)
        self.add_hotkey("KEY_ENTER", self.describe, "Describe", True)
        self.hotkey_display = HotkeyDisplay.opener(caller=self)
        self.refresh_data()

    def describe(self, _):
        """
        Describer callback for SingleRelationListers. As these list different datatypes in one control, the specific method to describe the entries is embedded
        in their controller data.
        """
        if "describer" not in self.selection.controller_data:
            Common.Session.set_message(
                "Resource cannot be described", Common.color("message_info")
            )
            return
        Common.Session.push_frame(
            self.selection.controller_data["describer"](
                entry=self.selection, entry_key="id"
            )
        )

    def get_data(self, *args, **kwargs):
        """
        Data retriever callback for SingleRelationListers. Retrieves the resources whose relations are to be discovered and parses each related resource out of
        that description.

        Yields
        ------
        awsc.termui.list_control.ListEntry
            A row to insert.
        """
        if self.descriptor is None:
            try:
                provider = Common.Session.service_provider(self.main_provider)
            except KeyError:
                Common.error(
                    "boto3 does not recognize provider",
                    "Invalid provider",
                    self.category,
                    subcategory=self.subcategory,
                    api_provider=self.main_provider,
                    classname=type(self).__name__,
                    source="SingleRelationLister",
                )
                return
            try:
                resp = getattr(provider, self.describe_method)(**self.describe_kwargs)
            except botoerror.ClientError as error:
                Common.clienterror(
                    error,
                    "Retrieve resource",
                    self.category,
                    subcategory=self.subcategory,
                    api_provider=self.main_provider,
                    api_method=self.describe_method,
                    api_kwargs=self.describe_kwargs,
                    classname=type(self).__name__,
                    source="SingleRelationLister",
                )
                return
            self.descriptor = (
                Common.Session.jq(self.object_path)
                .input(text=json.dumps(resp, default=datetime_hack))
                .first()
            )
            self.descriptor_raw = json.dumps(self.descriptor, default=datetime_hack)
        for elem in self.resource_descriptors:
            try:
                result = (
                    Common.Session.jq(elem["base_path"])
                    .input(text=self.descriptor_raw)
                    .first()
                )
                yield [
                    ListEntry(item, id=item, type=elem["type"], controller_data=elem)
                    for item in result
                    if item is not None
                ]
            except StopIteration:
                continue
            except ValueError:
                continue

    def refresh_data(self, *args, **kwargs):
        """
        Refresher. Initiates an asynchronous data refresh.
        """
        self.asynch(self.get_data, clear=True)

    def sort(self):
        self.entries.sort(key=Common.str_attrgetter("id"))
        self.entries.sort(key=Common.str_attrgetter("type"))
        self._cache = None


class MultiLister(ResourceListerBase):
    """
    A MulitLister is similar to a SingleRelationLister in the ability to list multiple resources in a single control. A MultiLister however acquires this data
    from several different listing API calls rather than from a single resource's description.
    """

    prefix = "CHANGEME"
    title = "CHANGEME"

    def title_info(self):
        """
        Title info functions are used to determine the title info after a lister has been instantiated.

        Returns
        -------
        str
            The information in the title.
        """

        return None

    @classmethod
    def opener(cls, **kwargs):
        """
        Session-aware initializer for this class. Creates an MultiLister object with the default alignment settings for the awsc
        layout.

        Returns
        -------
        list(awsc.base_control.MultiLister, awsc.info.HotkeyDisplay)
            A new instance of this class, and its associated hotkey display.
        """
        instance = cls(*default_args(), **default_kwargs(cls.prefix), **kwargs)
        instance.border = Common.default_border(
            cls.prefix, cls.title, instance.title_info()
        )
        return [instance, instance.hotkey_display]

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        compare_value,
        *args,
        compare_key="id",
        **kwargs,
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.add_column("type", 30, 0)
        self.add_column("id", 30, 1)
        if not hasattr(self, "resource_descriptors"):
            self.resource_descriptors = {}
            raise AttributeError("resource_descriptors is undefined")
        if isinstance(compare_value, ListEntry):
            self.compare_value = compare_value[compare_key]
            self.orig_compare_value = compare_value
        else:
            self.compare_value = compare_value
            self.orig_compare_value = compare_value
        self.add_hotkey("KEY_ENTER", self.describe, "Describe")
        self.add_hotkey("d", self.describe, "Describe")
        self.add_hotkey(ControlCodes.R, self.refresh_data, "Refresh")
        self.hotkey_display = HotkeyDisplay.opener(caller=self)
        self.refresh_data()

    def describe(self, *args):
        """
        Hotkey callback. Opens a generic describer with the data acquired from listing a specific entry as the browsed contents.
        """
        if self.selection is not None:
            Common.Session.push_frame(
                GenericDescriber.opener(
                    **{
                        "describing": self.selection["id"],
                        "content": json.dumps(
                            self.selection.controller_data, sort_keys=True, indent=2
                        ),
                        "pushed": True,
                    }
                )
            )

    def get_data(self, *args, **kwargs):
        """
        Data fetch callback for MultiListers. It's essentially the same as ResourceListers, except this one has several sets of data fetching calls defined.

        Yields
        ------
        awsc.termui.list_control.ListEntry
            A row to insert.
        """
        for elem in self.resource_descriptors:
            try:
                for y in self.get_data_generic(
                    elem["resource_key"],
                    elem["list_method"],
                    elem["list_kwargs"],
                    elem["item_path"],
                    elem["column_paths"],
                    elem["hidden_columns"],
                    None,
                    None,
                    elem,
                ):
                    yield y
            except NoResults:
                continue

    def matches(self, list_entry, *args):
        elem = args[0]
        raw_item = list_entry.controller_data
        if "compare_as_list" not in elem or not elem["compare_as_list"]:
            if callable(elem["compare_path"]):
                val = elem["compare_path"](raw_item)
            else:
                try:
                    val = (
                        Common.Session.jq(elem["compare_path"]).input(raw_item).first()
                    )
                except ValueError:
                    return False
                except StopIteration:
                    return False
            return val == self.compare_value
        if callable(elem["compare_path"]):
            return self.compare_value in elem["compare_path"](raw_item)

        try:
            for val in Common.Session.jq(elem["compare_path"]).input(raw_item).first():
                if val == self.compare_value:
                    return True
        except StopIteration:
            return False
        return False

    def refresh_data(self, *args, **kwargs):
        """
        Refresher. Initiates an asynchronous data refresh.
        """
        self.asynch(self.get_data, clear=True)

    def sort(self):
        self.entries.sort(key=attrgetter("type"))
        self._cache = None


class GenericDescriber(TextBrowser):
    """
    A generic describer control. GenericDescribers make no attempt to contact the AWS API. They're essentially session-aware text browsers.
    """

    prefix = "generic_describer"
    title = "Describe resource"

    @classmethod
    def opener(cls, **kwargs):
        """
        Session-aware initializer for this class. Creates an GenericDescriber object with the default alignment settings for the awsc
        layout.

        Returns
        -------
        list(awsc.base_control.GenericDescriber, awsc.info.HotkeyDisplay)
            A new instance of this class, and its associated hotkey display.
        """
        instance = cls(*default_args(), **default_kwargs(cls.prefix), **kwargs)
        instance.border = Common.default_border(
            cls.prefix, cls.title, instance.title_info()
        )
        return [instance, instance.hotkey_display]

    def title_info(self):
        """
        Title info functions are used to determine the title info after a lister has been instantiated.

        Returns
        -------
        str
            The information in the title.
        """
        return self.describing

    def __init__(
        self, parent, alignment, dimensions, describing, content, *args, **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.describing = describing
        self.add_text(content)
        self.hotkey_display = HotkeyDisplay.opener(caller=self)

    def toggle_wrap(self, *args, **kwargs):
        super().toggle_wrap(*args, **kwargs)
        Common.Session.set_message(
            f"Text wrap {'ON' if self.wrap else 'OFF'}",
            Common.color("message_info"),
        )

    @classmethod
    def register(cls):
        """
        Classmethod to register this class and all of its subclasses with the commander control.

        Also stores the class in the session control registry for other uses.

        This is why we can reach things via the command palette.
        """
        Common.Session.control_registry[cls.__name__] = RegistryHelper(
            opener=cls.opener
        )
        for subcls in cls.__subclasses__():
            subcls.register()


class DefaultResolver:
    """
    Resolver for resolving describer kwargs and entry attributes.

    The default resolver returns the value unchanged.
    """

    def __call__(self, value):
        return value


class AttributeResolver:
    """
    Resolver for resolving describer kwargs and entry attributes.

    Chainable resolver which returns an attribute of the passed value.
    """

    def __init__(self, attribute, resolver=DefaultResolver()):
        self.attribute = attribute
        self.resolver = resolver

    def __call__(self, value):
        return self.resolver(getattr(value, self.attribute))


class IndexResolver:
    """
    Resolver for resolving describer kwargs and entry attributes.

    Chainable resolver which returns an index of the passed value.
    """

    def __init__(self, index, resolver=DefaultResolver()):
        self.index = index
        self.resolver = resolver

    def __call__(self, value):
        return self.resolver(value[self.index])


class DescriberKwarg:
    """
    Helper class for describe_kwargs_override.
    """

    def __init__(self, kwarg, resolver=DefaultResolver()):
        self.kwarg = kwarg
        self.resolver = resolver

    def resolve(self, entry, **kwargs):
        """
        Resolves the runtime value of the describer kwarg based on values passed at init to the describer.
        """
        return self.resolver(kwargs[self.kwarg])


class DescriberEntryAttribute(DescriberKwarg):
    """
    Helper class for describe_kwargs_override.
    """

    def __init__(self, attribute, resolver=DefaultResolver()):
        super().__init__(
            "entry", resolver=AttributeResolver(attribute, resolver=resolver)
        )


class FString:
    def __init__(self, template):
        self.template = template

    def __call__(self, **kwargs):
        return self.template.format(**kwargs)


class Describer(TextBrowser):
    """
    Describers are AWS-aware text browsers. When provided with a resource ID, Describer subclasses contact the AWS API on a specified API endpoint, retrieve
    information about the resource, and display it or a subpath (depending on response format) in a syntax-highlighted, pretty-printed JSON form.
    """

    prefix = "CHANGEME"
    title = "CHANGEME"

    resource_type = "CHANGEME"
    main_provider = "CHANGEME"
    category = "CHANGEME"
    subcategory = "CHANGEME"
    describe_method = "CHANGEME"
    describe_kwarg_name: str | None = None
    describe_kwarg_is_list = False
    describe_kwargs_override: Dict[str, Any] = {}
    object_path: str | FString = ".ChangeMe"
    default_entry_key = "name"

    @classmethod
    def opener(cls, **kwargs):
        """
        Session-aware initializer for this class. Creates a Describer object with the default alignment settings for the awsc
        layout.

        Returns
        -------
        list(awsc.base_control.Describer, awsc.info.HotkeyDisplay)
            A new instance of this class, and its associated hotkey display.
        """
        instance = cls(
            *default_args(),
            **default_kwargs(cls.prefix),
            syntax_highlighting=True,
            scheme=Common.Configuration.scheme,
            **kwargs,
        )
        instance.border = Common.default_border(
            cls.prefix, cls.title, instance.title_info()
        )
        return [instance, instance.hotkey_display]

    def title_info(self):
        """
        Title info functions are used to determine the title info after a lister has been instantiated.

        Returns
        -------
        str
            The information in the title.
        """
        return self.entry_id

    def populate_entry(self, **kwargs):
        """
        Populates the entry fields of the Describer based on the entry and entry_key received.

        Expected to be called from the initializer.

        Parameters
        ----------
        entry : awsc.termui.list_control.ListEntry
            The list entry to populate with.
        entry_key : str
            The name of the primary key of the resource.
        """
        if kwargs["entry"] is None:
            self.entry = None
            self.entry_id = None
            return
        self.entry = kwargs["entry"]
        self.entry_id = kwargs["entry"][kwargs["entry_key"]]

    def populate_describe_kwargs(self, **kwargs):
        """
        Populates the describe kwargs dict, which is what will be sent to the AWS API when querying the resource.
        """
        if self.describe_kwarg_name is not None:
            self.describe_kwargs[self.describe_kwarg_name] = (
                [self.entry_id] if self.describe_kwarg_is_list else self.entry_id
            )
        for arg, arg_value in self.describe_kwargs_override.items():
            if hasattr(arg_value, "resolve"):
                self.describe_kwargs[arg] = arg_value.resolve(**kwargs)
            else:
                self.describe_kwargs[arg] = arg_value

    def confirm_template(self, *args, hotkey=None, hotkey_tooltip=None, **kwargs):
        """
        Creates a confirmation sequence.

        Can be used as a shorthand for registering a confirmation dialog as a hotkey. Returns a function that,
        when called, begins the confirmation process. The returned function requires no arguments, but takes any
        amount for legacy reasons.

        Args and kwargs correspond to the parameters of generic_confirm_template.

        Parameters
        ----------
        hotkey : str, optional
            The hotkey for this action. This string should be a valid key for the add_hotkey function. If not
            provided, no hotkey will be added for this confirmation dialog.
        hotkey_tooltip : str, optional
            If set, passes a tooltip for the hotkey to be displayed on the hotkey display list. If omitted, the
            hotkey will be hidden from the hotkey display list.

        Returns
        -------
        callable
            A function that, when called, begins the confirmation process.
        """

        kwargs["category"] = kwargs["category"] or self.category
        kwargs["subcategory"] = kwargs["subcategory"] or self.subcategory
        kwargs["resource_type"] = kwargs["resource_type"] or self.resource_type
        kwargs["provider"] = kwargs["provider"] or self.main_provider

        generic = generic_confirm_template(*args, **kwargs)

        def fn(*args):
            return generic(self, *args)

        if hotkey is not None:
            self.add_hotkey(hotkey, fn, hotkey_tooltip, True)
        return fn

    def __init__(
        self, parent, alignment, dimensions, *args, entry, entry_key=None, **kwargs
    ):
        if entry_key is None:
            entry_key = self.default_entry_key
        kwargs["entry"] = entry
        kwargs["entry_key"] = entry_key
        self.populate_entry(**kwargs)
        kwargs["entry_id"] = self.entry_id
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        if not hasattr(self, "describe_kwargs"):
            self.describe_kwargs = {}
        self.populate_describe_kwargs(**kwargs)

        self.add_hotkey(ControlCodes.R, self.refresh_data, "Refresh")
        self.hotkey_display = HotkeyDisplay.opener(caller=self)

        if hasattr(self, "additional_commands"):
            for key, command_spec in self.additional_commands.items():
                if "kwargs" not in command_spec:
                    command_spec[kwargs] = {}
                self.add_hotkey(
                    key,
                    self.command_wrapper(
                        command_spec["command"],
                        command_spec["data_arg"],
                        **command_spec["kwargs"],
                    ),
                    command_spec["tooltip"],
                )

        if isinstance(self.object_path, FString):
            self.object_path = self.object_path(**kwargs)

        self.refresh_data()

    def command(self, cmd, **kwargs):
        """
        Command caller. See awsc.base_control.ResourceLister.command.
        """
        frame = cmd(**kwargs)
        if frame is not None:
            Common.Session.push_frame(frame)

    def command_wrapper(self, cmd, data_arg, **kwargs):
        """
        Command wrapper. See awsc.base_control.ResourceLister.command_wrapper.
        """

        def fn(*args):
            kwargs[data_arg] = "\n".join(self.lines)
            self.command(
                cmd,
                pushed=True,
                caller=self,
                **kwargs,
            )

        return fn

    def toggle_wrap(self, *args, **kwargs):
        super().toggle_wrap(*args, **kwargs)
        Common.Session.set_message(
            f"Text wrap {'ON' if self.wrap else 'OFF'}",
            Common.color("message_info"),
        )

    def refresh_data(self, *args, **kwargs):
        """
        Data refresh callback. Fetches the info from the AWS API and refreshes the output.
        """
        try:
            provider = Common.Session.service_provider(self.main_provider)
        except KeyError:
            Common.error(
                "boto3 does not recognize provider",
                "Invalid provider",
                "Core",
                subcategory="Describer",
                api_provider=self.main_provider,
                classname=type(self).__name__,
            )
            return
        try:
            response = getattr(provider, self.describe_method)(**self.describe_kwargs)
        except botoerror.ClientError as error:
            Common.clienterror(
                error,
                "List Resources",
                "Core",
                subcategory="Describer",
                api_provider=self.main_provider,
                api_method=self.describe_method,
                api_kwargs=self.describe_kwargs,
            )
            return
        self.clear()
        self.add_text(
            json.dumps(
                Common.Session.jq(self.object_path)
                .input(text=json.dumps(response, default=datetime_hack))
                .first(),
                sort_keys=True,
                indent=2,
            )
        )

    @classmethod
    def register(cls):
        """
        Classmethod to register this class and all of its subclasses with the commander control.

        Also stores the class in the session control registry for other uses.

        This is why we can reach things via the command palette.
        """
        Common.Session.control_registry[cls.__name__] = RegistryHelper(
            opener=cls.opener
        )
        for subcls in cls.__subclasses__():
            subcls.register()


class DeleteResourceDialog(SessionAwareDialog):
    """
    Convenience dialog for requesting confirmation for an action. By default, this is tuned for delete actions but really, anything that requires a
    ok or cancel button can be put here.
    """

    line_size = 20

    def __init__(
        self,
        *args,
        resource_type,
        resource_identifier,
        callback,
        action_name="Delete",
        from_what=None,
        from_what_name=None,
        undoable=False,
        can_force=False,
        extra_fields=None,
        **kwargs,
    ):
        """
        Initializes a DeleteResourceDialog.

        Parameters
        ----------
        resource_type : str, optional
            The name of the type of resource being managed.
        resource_identifier : str, optional
            Identifier for the resource being managed. Does not need to be a primary key, this is for display
            purposes only.
        undoable : bool, default: False
            Cosmetic, to display a warning about data loss if not set to true.
        can_force : bool, default: False
            Adds an additional force checkbox to the confirmation dialog if set.
        extra_fields : dict, optional
            A map of additional fields to add to the confirmation dialog, where the key is the name of the field,
            and the value is a DialogField instance. These fields will be passed keyed by the key to the action callback.
        action_name : str, default: "Delete"
            The name of the action being performed if the dialog is confirmed. Defaults to 'Delete' as that is both
            the original purpose of confirmation dialogs and the most common action requiring a confirmation.
        from_what : str, optional
            If the action is to sever the link between two resources, we can add more context to the label by filling this field.
            This is the type of the other resource in the link if a link is being severed.
        from_what_name : str, optional
            This is the resource identifier for the resource named in from_what.
        """
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        kwargs["border"] = Common.default_border(
            "modal_dialog", f"{action_name} {resource_type}"
        )
        super().__init__(*args, **kwargs)
        label = [
            (f"{action_name} ", Common.color("modal_dialog_label")),
            (resource_type, Common.color("modal_dialog_label_highlight")),
            (' resource "', Common.color("modal_dialog_label")),
            (resource_identifier, Common.color("modal_dialog_label_highlight")),
            ('"', Common.color("modal_dialog_label")),
        ]

        if from_what is not None:
            label.append((f" from {from_what}", Common.color("modal_dialog_label")))
            if from_what_name is not None:
                label.append((' "', Common.color("modal_dialog_label")))
                label.append(
                    (from_what_name, Common.color("modal_dialog_label_highlight"))
                )
                label.append(('"', Common.color("modal_dialog_label")))

        label.append(("?", Common.color("modal_dialog_label")))
        self.set_title_label(label)
        if not undoable:
            self.add_field(
                DialogFieldLabel(
                    "This action cannot be undone.", Common.color("modal_dialog_error")
                )
            )
        self.highlighted = 1
        self.can_force = can_force
        if self.can_force:
            self.highlighted += 1
            self.force_checkbox = DialogFieldCheckbox(
                "Force",
                checked=False,
                color=Common.color("modal_dialog_textfield_label"),
                selected_color=Common.color("modal_dialog_textfield_selected"),
            )
            self.add_field(self.force_checkbox)
        self.extra_fields = {}
        if extra_fields is not None:
            for field in extra_fields.keys():
                self.highlighted += 1
                self.add_field(extra_fields[field])
                self.extra_fields[field] = extra_fields[field]
        self.callback = callback

    def input(self, key):
        if key.is_sequence and key.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(key)

    def accept_and_close(self):
        force = self.can_force and self.force_checkbox.checked
        self.callback(force=force, **self.extra_fields)
        self.close()


def format_timedelta(delta):
    """
    Timedelta formatter. Outputs a time difference in terms of the two highest time units for which it contains at least one.

    Parameters
    ----------
    delta : datetime.timedelta
        A time delta.

    Returns
    -------
    str
        An "x ago" representation of the delta.
    """
    hours = int(delta.seconds / 3600)
    minutes = int(delta.seconds / 60) - hours * 60
    seconds = delta.seconds - hours * 3600 - minutes * 60
    if delta.days > 0:
        return f"{delta.days}d{hours}h ago"
    if hours > 0:
        return f"{hours}h{minutes}m ago"
    if minutes > 0:
        return f"{minutes}m{seconds}s ago"
    if seconds > 0:
        return f"{seconds}s ago"
    return "<1s ago"


def isdumpable(obj):
    """
    Shorthand for checking if an object is a simple or a complex json type.

    Parameters
    ----------
    obj : object
        Any value.

    Returns
    -------
    bool
        True if obj is not a simple json type (int, float, bool or string).
    """
    return not isinstance(obj, (int, float, str, bool))


class ListResourceDocumentEditor:
    """
    Class for simplifying generating an editor for resources whose body is a JSON document.

    Attributes
    ----------
    provider : object
        Boto3 service provider.
    provider_name : str
        The name of the provider that was passed.
    static_fields : dict
        A mapping of field name to field value for keyword arguments that have a static value in the API call.
    """

    def __init__(
        self,
        provider,
        retrieve_method,
        retrieve_path,
        update_method,
        entry_name_arg,
        update_document_arg,
        entry_name_is_list=False,
        entry_key="name",
        entry_name_arg_update=None,
        as_json=True,
        message="Update successful",
        ignored_fields=None,
        static_fields=None,
    ):
        """
        Initializes a ListResourceDocumentEditor.

        Parameters
        ----------
        provider :str
            boto3 client provider name
        retrieve_method : str
            The method name for the provider to retrieve the document of the edited resource
        retrieve_path : str
            Path to the root of the document in the response for the retrieve method
        update_method : str
            The method name for the provider to call to update the resource
        entry_name_arg : str
            The name of the kwarg for the retrieve_method which must be passed the entry's key.
        update_document_arg : str
            The name of the kwarg for the update_method which receives the updated document.
        entry_name_is_list : bool
            If true, pass entry_name_arg to retrieve_method as a list.
        entry_key : str
            The field in the entry which contains the value for the entry name argument.
        entry_name_arg_update : str
            The name of the kwarg for the update_method which must be passed the entry's key.
            If None, use the same as for the retrieve_method.
        as_json : bool or dict
            If true, each field is passed as a json object. If false, all list and map fields are converted to their
            representation via json.dumps before passing. For a ListResourceDocumentEditor, this only affects the root object.
        ignored_fields : list
            A list of fields to ignore (remove) from the retrieved data. Informational fields that cause compatibility issues.
        static_fields : dict
            A mapping of field name to field value for keyword arguments that have a static value in the API call.
        """
        self.provider = Common.Session.service_provider(provider)
        self.provider_name = provider
        self.retrieve_method = retrieve_method
        if retrieve_method is not None:
            self.retrieve = getattr(self.provider, retrieve_method)
        if update_method is not None:
            self.update = getattr(self.provider, update_method)
        self.retrieve_path = retrieve_path
        self.retrieve_entry_name_arg = entry_name_arg
        self.retrieve_entry_name_is_list = entry_name_is_list
        self.update_document_arg = update_document_arg
        self.update_entry_name_arg = (
            entry_name_arg_update
            if entry_name_arg_update is not None
            else entry_name_arg
        )
        if ignored_fields is not None:
            self.ignored_fields = ignored_fields[:]
        else:
            self.ignored_fields = []
        self.entry_key = entry_key
        self.as_json = as_json
        self.update_method = update_method
        self.message = message
        self.static_fields = static_fields if static_fields is not None else {}

    def retrieve_content(self, selection):
        """
        Called when the data for the edited entry is being retrieved.

        Parameters
        ----------
        selection : awsc.termui.list_control.ListEntry
            A ListEntry or dict-like object which contains the primary key of the entry to edit at the set key.

        Returns
        -------
        str
            Return value of display_content on the fetched data.
        """
        r_kwargs = {
            self.retrieve_entry_name_arg: [selection[self.entry_key]]
            if self.retrieve_entry_name_is_list
            else selection[self.entry_key]
        }
        response = self.retrieve(**r_kwargs)
        return self.display_content(response)

    def display_content(self, response):
        """
        Creates the display format for the resource retrieved from AWS.

        Parameters
        ----------
        response : dict
            The result of an AWS API call.

        Returns
        -------
        str
            A prettified JSON, which the user will be editing.
        """
        inner = (
            Common.Session.jq(self.retrieve_path)
            .input(text=json.dumps(response, default=datetime_hack))
            .first()
        )
        for field in self.ignored_fields:
            if field in inner:
                del inner[field]
        for field, field_value in self.static_fields.items():
            inner[field] = field_value
        return json.dumps(
            inner,
            sort_keys=True,
            indent=2,
        )

    def update_content(self, selection, newcontent):
        """
        Updates the data on AWS with the data edited by the user.

        Parameters
        ----------
        selection : awsc.termui.list_control.ListEntry
            A ListEntry or dict-like object which contains the primary key of the entry to edit at the set key.
        newcontent : str
            The data as edited by the user.
        """
        try:
            u_kwargs = (
                {
                    self.update_entry_name_arg: selection[self.entry_key],
                    self.update_document_arg: json.loads(newcontent)
                    if self.as_json
                    else newcontent,
                }
                if self.update_document_arg is not None
                else json.loads(newcontent)
            )
            self.update(**u_kwargs)
            Common.success(
                self.message,
                self.update_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.update_method,
            )
        except json.JSONDecodeError as error:
            Common.error(
                f"JSON decode error: {error}",
                self.update_method,
                self.provider_name.capitalize(),
                raw=newcontent,
            )
        except botoerror.ClientError as error:
            Common.clienterror(
                error,
                self.update_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.update_method,
            )
        except botoerror.ParamValidationError as error:
            Common.error(
                f"Parameter validation error: {str(error)}",
                self.update_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.update_method,
            )

    def edit(self, selection):
        """
        Executes the full editing workflow. Retrieves the data from AWS, presents it to the user to edit, then calls back to AWS to update the content.
        """
        content = self.retrieve_content(selection)
        newcontent = Common.Session.textedit(content)
        if content == newcontent:
            Common.Session.set_message("Input unchanged.", Common.color("message_info"))
            return
        self.update_content(selection, newcontent.strip(" \n\t"))

    def hotkey_callback(self, control):
        """
        This function can be passed as a hotkey callback to a ResourceLister.
        """
        self.edit(control.selection)


class ListResourceDocumentCreator(ListResourceDocumentEditor):
    """
    Class for simplifying generating an editor for resources whose body is a JSON document. This flavour is specifically tuned for create operations.

    Attributes
    ----------
    golden : str
        The initial document in string form. Nothing happens if the state of the document when exiting the editor matches the initial document.
    static_fields : dict
        A mapping of field name to field value for keyword arguments that have a static value in the API call.
    """

    def __init__(
        self,
        provider,
        create_method,
        create_document_arg,
        initial_document=None,
        as_json=True,
        static_fields=None,
        message="Create successful",
    ):
        """
        Initializes a ListResourceDocumentCreator.

        Parameters
        ----------
            provider : str
                boto3 client provider name
            create_method : str
                The method name for the provider to call to create the resource
            create_document_arg : str
                The name of the kwarg for the create_method which receives the updated document.
                If empty, the root level object is passed as a set of kwargs.
            initial_document : dict
                The initial document to present for editing. Consider it a template of fields to be filled.
                Expected to be a dict or dict-like structure that can translate to a json string via json.dumps.
            as_json : bool or list, default=True
                Whether certain fields need to be passed as json or json string. Unfortunately, this is not uniform across
                the AWS APIs. The behaviour is that ints, floats, bools and strings are never changed by this controller.
                Dicts and lists, by default - if this value is True - are dumped as strings. You can disable this behaviour
                by setting this field to False. You can also selectively choose a set of fields where this behaviour should be
                enabled, by setting this to a list of field names. And yes, unfortunately this is a valid use case, there are
                AWS APIs where certain fields are strings of json, while others are json, I'm looking at you, policies.
            static_fields : dict
                A mapping of field name to field value for keyword arguments that have a static value in the API call.
        """
        super().__init__(
            provider,
            None,
            ".",
            create_method,
            None,
            create_document_arg,
            entry_key=None,
            entry_name_arg_update=None,
            as_json=as_json,
            message=message,
            static_fields=static_fields,
        )
        self.initial_document = initial_document
        self.golden = self.display_content(self.initial_document)
        self.create_method = create_method
        self.provider_name = provider

    def retrieve_content(self, selection):
        from copy import deepcopy

        return self.display_content(deepcopy(self.initial_document))

    def generate_kwargs(self, selection, newcontent):
        """
        Generates the kwargs for updating based on the kwargs settings within the object instance.

        Parameters
        ----------
        selection : awsc.termui.list_control.ListEntry
            A ListEntry or dict-like object which contains the primary key of the entry to edit at the set key.
        newcontent : str
            The data as edited by the user.
        """
        update_data = json.loads(newcontent)
        if self.update_document_arg is not None:
            u_kwargs = {
                self.update_document_arg: update_data if self.as_json else newcontent
            }
        else:
            u_kwargs = {}
            for field in update_data.keys():
                if (
                    not isdumpable(update_data[field])
                    or self.as_json is True
                    or (isinstance(self.as_json, list) and field in self.as_json)
                ):
                    u_kwargs[field] = update_data[field]
                else:
                    u_kwargs[field] = json.dumps(
                        update_data[field], default=datetime_hack
                    )
        for field in self.static_fields.keys():
            u_kwargs[field] = self.static_fields[field]
        return u_kwargs

    def update_content(self, selection, newcontent):
        if newcontent == self.golden:
            Common.Session.set_message("Cancelled", Common.color("message_info"))
            return
        try:
            u_kwargs = self.generate_kwargs(selection, newcontent)
        except json.JSONDecodeError as error:
            Common.Session.set_message(
                f"JSON parse error: {error}", Common.color("message_error")
            )
            return
        try:
            self.update(**u_kwargs)
            Common.success(
                self.message,
                self.create_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.create_method,
            )
        except botoerror.ClientError as error:
            Common.clienterror(
                error,
                self.create_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.create_method,
            )
        except botoerror.ParamValidationError as error:
            Common.error(
                f"Parameter validation error: {str(error)}",
                self.create_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.create_method,
            )

    def edit(self, selection=None):
        super().edit(selection)


class ListResourceFieldsEditor(ListResourceDocumentEditor):
    """
    Class for simplifying generating an editor for resources which are essentially just a set of fields. Allows for a more restricted editing session compared
    to ListResourceDocumentEditor.
    """

    def __init__(
        self,
        provider,
        retrieve_method,
        retrieve_path,
        update_method,
        entry_name_arg,
        fields,
        entry_key="name",
        entry_name_arg_update=None,
        as_json=True,
        message="Update successful",
    ):
        super().__init__(
            provider,
            retrieve_method,
            retrieve_path,
            update_method,
            entry_name_arg,
            None,
            entry_key="name",
            entry_name_arg_update=entry_name_arg_update,
            as_json=as_json,
            message=message,
        )

        self.fields = fields
        self.orig = {}

    def display_content(self, response):
        content = json.dumps(response, default=datetime_hack)
        ret = {}
        for field, path in self.fields.items():
            ret[field] = Common.Session.jq(path).input(text=content).first()
            self.orig[field] = ret[field]
        return json.dumps(ret, sort_keys=True, indent=2)

    def update_content(self, selection, newcontent):
        update_data = json.loads(newcontent)
        u_kwargs = {self.update_entry_name_arg: selection[self.entry_key]}
        for field in self.fields.keys():
            if self.orig[field] == update_data[field]:
                continue
            if (
                not isdumpable(update_data[field])
                or self.as_json is True
                or (isinstance(self.as_json, list) and field in self.as_json)
            ):
                u_kwargs[field] = update_data[field]
            else:
                u_kwargs[field] = json.dumps(update_data[field], default=datetime_hack)
        try:
            self.update(**u_kwargs)
            Common.Session.set_message(
                "Update successful", Common.color("message_success")
            )
        except botoerror.ClientError as error:
            Common.clienterror(
                error,
                self.update_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.update_method,
            )
        except botoerror.ParamValidationError as error:
            Common.error(
                f"Parameter validation error: {str(error)}",
                self.update_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.update_method,
            )



"""
Module for EKS clusters.
"""
import subprocess
from pathlib import Path

import yaml

from .arn import ARN
from .base_control import Describer, ResourceLister
from .common import Common


class EKSDescriber(Describer):
    """
    Describer control for EKS clusters.
    """

    prefix = "eks_browser"
    title = "EKS Cluster"

    resource_type = "eks cluster"
    main_provider = "eks"
    category = "EKS"
    subcategory = "Clusters"
    describe_method = "describe_cluster"
    describe_kwarg_name = "name"
    object_path = ".cluster"


class EKSResourceLister(ResourceLister):
    """
    Lister control for EKS clusters.
    """

    prefix = "eks_list"
    title = "EKS Clusters"
    command_palette = ["eks", "kubernetes"]

    resource_type = "eks cluster"
    main_provider = "eks"
    category = "EKS"
    subcategory = "Clusters"
    list_method = "list_clusters"
    item_path = ".clusters"
    columns = {
        "name": {
            "path": ".",
            "size": 100,
            "weight": 1,
            "sort_weight": 1,
        }
    }
    describe_command = EKSDescriber.opener

    @ResourceLister.Autohotkey("k", "Fetch kube context", True)
    def fetch_kubecontext(self, _):
        kc_path = Path.home() / ".kube" / "config"
        if kc_path.exists():
            with kc_path.open("r") as file:
                kubeconfig = yaml.safe_load(file.read())
        else:
            kubeconfig = {
                "apiVersion": "v1",
                "kind": "Config",
                "clusters": [],
                "contexts": [],
                "users": [],
                "preferences": {},
            }
        resp = Common.generic_api_call(
            "eks",
            "describe_cluster",
            {"name": self.selection["name"]},
            "Describe EKS cluster",
            "EKS",
            subcategory="Cluster",
            resource=self.selection["name"],
        )
        if resp["Success"]:
            cluster_data = resp["Response"]
            endpoint = cluster_data["cluster"]["endpoint"]
            ca_data = cluster_data["cluster"]["certificateAuthority"]["data"]
            cluster_arn = cluster_data["cluster"]["arn"]
            cluster_arn_parsed = ARN(cluster_arn)

            idx = -1
            for cidx, elem in enumerate(kubeconfig["clusters"]):
                if elem["name"] == cluster_arn:
                    idx = cidx
                    break
            cluster = {"certificate-authority-data": ca_data, "server": endpoint}
            if idx == -1:
                obj = {"cluster": cluster, "name": cluster_arn}
                kubeconfig["clusters"].append(obj)
            else:
                kubeconfig["clusters"][idx]["cluster"] = cluster

            idx = -1
            for cidx, elem in enumerate(kubeconfig["contexts"]):
                if elem["name"] == cluster_arn:
                    idx = cidx
                    break
            context = {"cluster": cluster_arn, "user": cluster_arn}
            if idx == -1:
                obj = {"context": context, "name": cluster_arn}
                kubeconfig["contexts"].append(obj)
            else:
                kubeconfig["contexts"][idx]["context"] = context

            idx = -1
            for cidx, elem in enumerate(kubeconfig["users"]):
                if elem["name"] == cluster_arn:
                    idx = cidx
                    break
            user = {
                "exec": {
                    "apiVersion": "client.authentication.k8s.io/v1beta1",
                    "command": "aws",
                    "args": [
                        "--region",
                        cluster_arn_parsed.region,
                        "eks",
                        "get-token",
                        "--cluster-name",
                        cluster_arn_parsed.resource_id_first,
                        "--profile",
                        Common.Session.context,
                    ],
                }
            }
            if idx == -1:
                obj = {"user": user, "name": cluster_arn}
                kubeconfig["users"].append(obj)
            else:
                kubeconfig["users"][idx]["user"] = user

            kubeconfig["current-context"] = cluster_arn

            with kc_path.open("w", encoding="utf-8") as file:
                file.write(yaml.dump(kubeconfig))

            Common.Session.set_message(
                f"Successfully wrote configuration {cluster_arn} to ~/.kube/config.",
                Common.color("message_success"),
            )
            return True
        return False

    @ResourceLister.Autohotkey("9", "k9s", True)
    def open_k9s(self, _):
        if self.fetch_kubecontext(_):
            exit_code = Common.Session.ui.unraw(
                subprocess.run,
                [
                    "k9s",
                ],
            )
            Common.Session.set_message(
                f"k9s exited with code {exit_code.returncode}",
                Common.color("message_info"),
            )



"""
Module for common elements for IAM resources.
"""
from copy import deepcopy

from botocore import exceptions as botoerror

from .base_control import (
    Describer,
    ListResourceDocumentCreator,
    ResourceLister,
    SelectionAttribute,
    SingleNameDialog,
    TemplateDict,
)
from .common import Common
from .termui.common import Commons
from .termui.ui import ControlCodes

EMPTY_POLICY_DOCUMENT = {
    "PolicyDocument": {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Deny",
                "Action": [""],
                "Resource": "*",
                "Condition": {},
            }
        ],
    }
}
"""
Anatomy of an empty policy document.
"""


class InlinePolicyAttacher:
    """
    Helper class for attaching inline policies to the big 3 IAM types - users, groups and roles.

    Attributes
    ----------
    datatype : str
        The type of data (user, role or group).
    parent : awsc.termui.control.Control
        The control using the attacher.
    client : object
        Boto3 service provider.
    get : callable
        Method reference to the datatype policy getter of self.client.
    put_method : str
        Method name for the datatype policy setter of self.client.
    arg : str
        The argname for the put_method.
    selection_is_policy : bool, default=False
        Whether the parent is listing policies rather than resources that receive policies.
    """

    def __init__(self, parent, datatype, selection_is_policy=False):
        self.datatype = datatype
        self.parent = parent
        self.client = Common.Session.service_provider("iam")

        self.get = getattr(self.client, f"get_{datatype}_policy")
        self.put_method = f"put_{datatype}_policy"
        self.arg = f"{datatype.capitalize()}Name"

        self.selection_is_policy = selection_is_policy

        if not selection_is_policy:
            self.parent.add_hotkey(
                "i", self.attach_inline_policy, "Put inline policy", True
            )

    def attach_inline_policy(self, _):
        """
        Hotkey callback for attaching an inline policy to the resource listed by the parent.
        """
        SingleNameDialog(
            self.parent.parent,
            f"Attach new or edit existing inline policy for {self.datatype}",
            self.do_attach_inline,
            label="Policy name:",
            what="name",
            subject="policy",
            default="",
            caller=self.parent,
            accepted_inputs=Commons.alphanum_with_symbols("+=,.@_-"),
        )

    def do_attach_inline(self, name):
        """
        Action callback for attaching an inline policy.

        Parameters
        ----------
        name : str
            The name of the resource not managed by the lister parent. This is a policy name if self.selection_is_policy is False,
            otherwise a role/user/group name.
        """
        if self.parent.selection is None:
            return
        if self.selection_is_policy:
            kwargs = {"PolicyName": self.parent.selection["name"], self.arg: name}
        else:
            kwargs = {self.arg: self.parent.selection["name"], "PolicyName": name}
        try:
            chk = self.get(**kwargs)
            initial_document = {"PolicyDocument": chk["PolicyDocument"]}
        except self.client.exceptions.NoSuchEntityException:
            initial_document = deepcopy(EMPTY_POLICY_DOCUMENT)
        except botoerror.ClientError as error:
            Common.clienterror(
                error,
                f"Retrieve Inline {self.datatype.capitalize()} Policy",
                "IAM",
                set_message=True,
            )
            return
        except Exception as error:
            Common.error(
                str(error),
                f"Retrieve Inline {self.datatype.capitalize()} Policy",
                "IAM",
                set_message=True,
            )
            return

        creator = ListResourceDocumentCreator(
            "iam",
            self.put_method,
            None,
            initial_document,
            as_json=False,
            static_fields=kwargs,
            message="Put policy successful",
        )
        creator.edit()
        self.parent.refresh_data()


class InlinePolicyDescriber(Describer):
    """
    Describer control for inline policies.
    """

    prefix = "user_inline_policy_browser"
    title = "Inline Policy"

    resource_type = "inline policy"
    main_provider = "iam"
    category = "IAM"
    subcategory = "Inline Policies"
    object_path = "."

    def __init__(self, *args, caller=None, datatype="", data=None, **kwargs):
        self.describe_method = f"get_{datatype}_policy"
        self.arg = f"{datatype.capitalize()}Name"
        self.describe_kwarg_name = "PolicyName"
        self.describe_kwargs = {self.arg: data["name"]}
        super().__init__(*args, caller=caller, **kwargs)


@ResourceLister.Autocommand(
    "UserLister", "p", "View inline policies", "data", datatype="user"
)
@ResourceLister.Autocommand(
    "GroupLister", "p", "View inline policies", "data", datatype="group"
)
@ResourceLister.Autocommand(
    "RoleLister", "p", "View inline policies", "data", datatype="role"
)
class InlinePolicyLister(ResourceLister):
    """
    Lister control for inline policies.
    """

    prefix = "inline_policy_list"
    title = "Inline Policies"

    resource_type = "inline policy"
    main_provider = "iam"
    category = "IAM"
    subcategory = "Inline Policies"
    columns = {
        "name": {
            "path": ".",
            "size": 25,
            "weight": 0,
            "sort_weight": 0,
        },
    }
    describe_command = InlinePolicyDescriber.opener

    def title_info(self):
        return f"{self.datatype.capitalize()}: {self.data['name']}"

    @ResourceLister.Autohotkey("e", "Edit Policy", True)
    def edit_policy(self, _):
        """
        Hotkey callback for editing an inline policy.
        """
        self.attacher.do_attach_inline(self.data["name"])

    def __init__(self, *args, datatype="", data=None, **kwargs):
        self.data = data
        self.datatype = datatype
        self.attacher = InlinePolicyAttacher(self, datatype, selection_is_policy=True)
        self.arg = f"{datatype.capitalize()}Name"
        self.resource_key = "iam"
        self.delete_method = f"delete_{datatype}_policy"
        self.list_method = f"list_{datatype}_policies"
        self.item_path = ".PolicyNames"
        self.list_kwargs = {self.arg: self.data["name"]}
        self.column_paths = {
            "name": ".",
        }
        self.imported_column_sizes = {
            "name": 25,
        }
        self.describe_command = InlinePolicyDescriber.opener
        self.describe_kwargs = {"datatype": datatype, "data": data}

        self.imported_column_order = ["name"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)
        self.confirm_template(
            self.delete_method,
            TemplateDict(
                {"PolicyName": SelectionAttribute("name"), self.arg: self.data["name"]}
            ),
            hotkey=ControlCodes.D,
            hotkey_tooltip="Delete Policy",
        )



"""
Definition snippets that may be used by other code.
"""
from copy import deepcopy

from .base_control import tag_finder_generator

MULTILISTER_DESCRIPTORS = {
    "ec2": {
        "resource_key": "ec2",
        "list_method": "describe_instances",
        "list_kwargs": {},
        "item_path": "[.Reservations[].Instances[]]",
        "column_paths": {
            "type": lambda x: "EC2 Instance",
            "id": ".InstanceId",
            "name": tag_finder_generator("Name"),
        },
        "hidden_columns": {},
    },
    "rds": {
        "resource_key": "rds",
        "list_method": "describe_db_instances",
        "list_kwargs": {},
        "item_path": ".DBInstances",
        "column_paths": {
            "type": lambda x: "RDS Instance",
            "id": ".DBInstanceIdentifier",
            "name": ".Endpoint.Address",
        },
        "hidden_columns": {},
    },
}
"""
Commonly used multilister descriptors.
"""


def multilister_with_compare_path(resource_key, compare_path, compare_as_list=False):
    """
    Convenience function for setting the compare path of a common multilister descriptor.

    Parameters
    ----------
    resource_key : str
        Key in MULTILISTER_DESCRIPTORS.
    compare_path : str
        The compare path to set.
    compare_as_list : bool
        The value to set for compare_as_list.

    Returns
    -------
    dict
        The appropriate key from MULTILISTER_DESCRIPTORS with compare_path set.
    """
    ret = deepcopy(MULTILISTER_DESCRIPTORS[resource_key])
    ret["compare_path"] = compare_path
    ret["compare_as_list"] = compare_as_list
    return ret



"""
Module which imports all the resource controls and registers an init hook to register the controls with the session.
"""
from .base_control import Describer, GenericDescriber, ResourceLister
from .common import Common
from .dashboard import Dashboard
from .resource_ami import *
from .resource_asg import *
from .resource_beanstalk import *
from .resource_cfn import *
from .resource_cloudwatch import *
from .resource_dsg import *
from .resource_ebs import *
from .resource_ec2 import *
from .resource_ec2_class import *
from .resource_eks import *
from .resource_iam import *
from .resource_iam_group import *
from .resource_iam_instance_profile import *
from .resource_iam_policy import *
from .resource_iam_role import *
from .resource_iam_user import *
from .resource_keypair import *
from .resource_lb import *
from .resource_lc import *
from .resource_listener import *
from .resource_r53 import *
from .resource_rds import *
from .resource_routing import *
from .resource_s3 import *
from .resource_sg import *
from .resource_sqs import *
from .resource_subnet import *
from .resource_tg import *
from .resource_vpc import *

Common.run_on_init(ResourceLister.register)
Common.run_on_init(Describer.register)
Common.run_on_init(GenericDescriber.register)
Common.run_on_init(Dashboard.register)



"""
This module defines dialogs, which are modal controls designed to be displayed as topmost blocks for inputting data.
"""
from .color import ColorBlackOnGold, ColorBlackOnOrange, ColorGold
from .common import Commons
from .control import Control


class DialogField:
    """
    Base class for input fields on a dialog.

    Attributes
    ----------
    highlightable : bool
        Whether the field can be selected. Unhighlightable fields are useful for labels.
    centered : bool
        Whether to forcibly center the field within the dialog.
    """

    def __init__(self):
        self.highlightable = False
        self.centered = False
        Commons.UIInstance.dirty = True

    def input(self, inkey):
        """
        Input handler method for the field. Works identically to the input method of a block.

        Parameters
        ----------
        inkey : blessed.keyboard.Keystroke
            The key that was pressed.

        Returns
        -------
        bool
            Whether the input was handled.
        """
        return False

    def paint(self, x0, x1, y, selected=False):
        """
        Paint callback for the field. Responsible for outputting the entire field.

        Parameters
        ----------
        x0 : int
            The left bound of the field, as determined by the parent dialog.
        x1 : int
            The right bound of the field, as determined by the parent dialog.
        y : int
            The y of the field, as determined by the parent dialog.
        selected : bool, default=False
            Whether the field is currently selected.
        """

    @property
    def value(self):
        """
        Read-only property for retrieving a generalized value of the field.
        """
        return None


class DialogFieldLabel(DialogField):
    """
    A field which displays a static text. Cannot be selected.

    Attributes
    ----------
    default_color : awsc.termui.color.Color
        The default color for the text of this label.
    texts : list(tuple(str, awsc.termui.color.Color))
        A list of text snippets, each associated with a color to draw it with. Text snippets are printed sequentially.
        If any list entry is a str instead of a tuple, it will be printed with the default color. This is done for the sake of both legacy
        and simplicity.
    """

    def __init__(self, texts, default_color=ColorGold, centered=True):
        super().__init__()
        self.default_color = default_color
        if isinstance(texts, str):
            self.texts = [(texts, default_color)]
        else:
            self.texts = list(texts[:])
        self.centered = centered

    @property
    def value(self):
        return self.text

    @property
    def text(self):
        """
        Property. Returns the raw text displayed by the label as a string. When set, replaces the entire text of the field with
        the provided text and default color.
        """
        return "".join([a[0] for a in self.texts])

    @text.setter
    def text(self, value):
        if isinstance(value, list):
            self.texts = value
        else:
            self.texts = [(value, self.default_color)]

    def add_text(self, text, color=None):
        """
        Adds a new text snippet to the label, optionally with a set color.

        Parameters
        ----------
        text : str
            The text snippet to add.
        color : awsc.termui.color.Color, optional
            The color of the text. Default color attribute is used if omitted.
        """
        if color is None:
            color = self.default_color
        self.texts.append((text, color))

    def paint(self, x0, x1, y, selected=False):
        x = x0
        if self.centered:
            textlen = 0
            for text in self.texts:
                textlen += len(text[0])
            width = x1 - x0 + 1
            x = int(width / 2) - int(textlen / 2) + x0 + 1
        for text in self.texts:
            Commons.UIInstance.print(text[0], xy=(x, y), color=text[1])
            x += len(text[0])


class DialogFieldButton(DialogField):
    """
    A button on the dialog. While selected, hitting the enter key will execute the button's callback function.

    Attributes
    ----------
    text : str
        The text to display on the button.
    color : awsc.termui.color.Color
        The color of the button when not selected.
    selected_color : awsc.termui.color.Color
        The color of the button when selected.
    action : callable
        The callback to call when the button is pressed. No parameters are passed to the callback.
    """

    def __init__(self, text, action, color=ColorGold, selected_color=ColorBlackOnGold):
        super().__init__()
        self.highlightable = True
        self.text = text
        self.color = color
        self.selected_color = selected_color
        self.centered = True
        self.action = action

    @property
    def value(self):
        return self.text

    def input(self, inkey):
        if inkey.is_sequence and inkey.name == "KEY_ENTER":
            self.action()
            Commons.UIInstance.dirty = True
        return True

    def paint(self, x0, x1, y, selected=False):
        x = x0
        if self.centered:
            textlen = len(self.text) + 4
            width = x1 - x0 + 1
            x = int(width / 2) - int(textlen / 2) + x0
        Commons.UIInstance.print(
            f"< {self.text} >",
            xy=(x, y),
            color=self.selected_color if selected else self.color,
        )


class DialogFieldCheckbox(DialogField):
    """
    A checkbox toggle on the dialog. While selected, hitting the enter key will toggle the checked state of the checkbox.

    Attributes
    ----------
    label : str
        The text to display next to the checkbox.
    color : awsc.termui.color.Color
        The color of the checkbox when not selected.
    selected_color : awsc.termui.color.Color
        The color of the checkbox when selected.
    checked : bool
        Whether the checkbox is checked.
    char_unchecked : str
        The checkbox display character for when it is not checked.
    char_checked : str
        The checkbox display character for when it is checked.
    """

    def __init__(
        self, label, checked=False, color=ColorGold, selected_color=ColorBlackOnGold
    ):
        super().__init__()
        self.highlightable = True
        self.checked = checked
        self.label = label
        self.color = color
        self.selected_color = selected_color
        self.char_unchecked = "☐"
        self.char_checked = "☑"

    @property
    def value(self):
        return self.checked

    def input(self, inkey):
        if inkey.is_sequence and inkey.name == "KEY_ENTER":
            self.checked = not self.checked
            Commons.UIInstance.dirty = True
        return True

    def paint(self, x0, x1, y, selected=False):
        x = x0
        checkbox_character = self.char_checked if self.checked else self.char_unchecked
        text = f"{checkbox_character} {self.label}"
        if self.centered:
            textlen = len(text)
            width = x1 - x0 + 1
            x = int(width / 2) - int(textlen / 2) + x0
        Commons.UIInstance.print(
            text, xy=(x, y), color=self.selected_color if selected else self.color
        )


class DialogFieldText(DialogField):
    """
    An input textfield on the dialog.

    Attributes
    ----------
    label : str
        The text to display next to the input field.
    text : str
        The text that is currently typed into the input field.
    color : awsc.termui.color.Color
        The color of the textfield when not selected.
    selected_color : awsc.termui.color.Color
        The color of the textfield when selected.
    label_color : awsc.termui.color.Color
        The color of the label of the textfield.
    label_min : int
        The minimum width that must be allocated for the label.
    password : bool
        If set, masks the text of the textfield with asterisks.
    drawable : int
        The calculated width of the interactive part of the textfield.
    left : int
        How much the textfield is scrolled to the right, in characters.
    accepted_inputs : str
        A string of characters which are acceptable inputs in the textfield. No character not present in the string will be typeable into
        the textfield.
    """

    def __init__(
        self,
        label,
        text="",
        color=ColorBlackOnOrange,
        selected_color=ColorBlackOnGold,
        label_color=ColorGold,
        label_min=0,
        password=False,
        accepted_inputs=None,
    ):
        super().__init__()
        self.highlightable = True
        self.left = 0
        self.text = text
        self.label = label
        self.color = color
        self.label_color = label_color
        self.label_min = label_min
        self.selected_color = selected_color
        self.centered = True
        self.password = password
        self.drawable = 0
        if accepted_inputs is None:
            self.accepted_inputs = Commons.TextfieldInputs
        else:
            self.accepted_inputs = accepted_inputs

    @property
    def value(self):
        return self.text

    def input(self, inkey):
        if inkey.is_sequence:
            if inkey.name == "KEY_LEFT":
                if self.left > 0:
                    self.left -= 1
                Commons.UIInstance.dirty = True
                return True
            if inkey.name == "KEY_RIGHT":
                self.left += 1
                Commons.UIInstance.dirty = True
                return True
            if inkey.name == "KEY_HOME":
                self.left = 0
                Commons.UIInstance.dirty = True
                return True
            if inkey.name == "KEY_END":
                self.left = max(len(self.text) - self.drawable, 0)
                Commons.UIInstance.dirty = True
                return True
            if inkey.name == "KEY_BACKSPACE" and len(self.text) > 0:
                self.text = self.text[:-1]
                Commons.UIInstance.dirty = True
                return True
            if inkey.name == "KEY_DELETE":
                self.text = ""
                Commons.UIInstance.dirty = True
                return True
        elif inkey in self.accepted_inputs:
            self.text += inkey
            if len(self.text) > self.drawable:
                self.left = len(self.text) - self.drawable + 1
            Commons.UIInstance.dirty = True
            return True
        return False

    def paint(self, x0, x1, y, selected=False):
        x = x0
        Commons.UIInstance.print(self.label, xy=(x, y), color=self.label_color)
        x += max(len(self.label) + 1, self.label_min)
        space = x1 - x + 1
        self.drawable = space
        if self.left >= len(self.text):
            self.left = 0
        text = self.text[
            self.left : (
                self.left + space
                if len(self.text) > self.left + space
                else len(self.text)
            )
        ]
        if self.password:
            text = "*" * len(text)
        if len(text) < space:
            text += " " * (space - len(text))
        Commons.UIInstance.print(
            text, xy=(x, y), color=self.selected_color if selected else self.color
        )


class DialogControl(Control):
    """
    A generic dialog. A dialog is intended to be used as a form that can be filled by the user for some action to happen.

    Attributes
    ----------
    highlighted : int
        The index of the control that is currently selected.
    fields : list(awsc.termui.dialog.DialogField)
        A list of field objects associated with the dialog.
    bookmark : int
        The position where the next field will be inserted. The confirm and cancel buttons, if used, are forcibly pushed as the last controls
        in the list, so bookmark always preceeds the confirm and cancel buttons.
    background_color : awsc.termui.color.Color
        The background color to fill the dialog with.
    """

    def __init__(
        self,
        *args,
        ok_action=None,
        cancel_action=None,
        background_color=None,
        confirm_text="Ok",
        cancel_text="Cancel",
        color=ColorGold,
        selected_color=ColorBlackOnGold,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.highlighted = 0
        self.bookmark = 0
        self.background_color = background_color

        self.fields = []
        if confirm_text is not None:
            self.fields.append(
                DialogFieldButton(
                    confirm_text,
                    ok_action if ok_action is not None else self.nop,
                    color,
                    selected_color,
                )
            )
        if cancel_text is not None:
            self.fields.append(
                DialogFieldButton(
                    cancel_text,
                    cancel_action if cancel_action is not None else self.nop,
                    color,
                    selected_color,
                )
            )

    def nop(self):
        """
        A callable that does nothing, for the purpose of having such a callable.
        """

    def add_field(self, field):
        """
        Inserts a new field into the dialog. This will be inserted as the last field before the confirm and cancel buttons.

        Parameters
        ----------
        field : awsc.termui.dialog.DialogField
            The field to insert.
        """
        self.fields.insert(self.bookmark, field)
        self.bookmark += 1
        for i, elem in enumerate(self.fields):
            if elem.highlightable:
                self.highlighted = i
                break

    def input(self, key):
        if key == "\t":
            orig = self.highlighted
            while True:
                self.highlighted += 1
                if self.highlighted >= len(self.fields):
                    self.highlighted -= len(self.fields)
                if self.fields[self.highlighted].highlightable:
                    break
                if self.highlighted == orig:
                    break
            Commons.UIInstance.dirty = True
            return True
        if key.is_sequence and key.name == "KEY_BTAB":
            orig = self.highlighted
            while True:
                self.highlighted -= 1
                if self.highlighted < 0:
                    self.highlighted += len(self.fields)
                if self.fields[self.highlighted].highlightable:
                    break
                if self.highlighted == orig:
                    break
            Commons.UIInstance.dirty = True
            return True
        self.fields[self.highlighted].input(key)
        return True  # Modals should prevent input from being piped to others.

    def paint(self):
        corners = self.corners
        if self.background_color is not None:
            width = corners[0][1] - corners[0][0] + 1
            for row in range(corners[1][0], corners[1][1] + 1):
                Commons.UIInstance.print(
                    " " * width, xy=(corners[0][0], row), color=self.background_color
                )
        super().paint()
        buttons = []
        indices = {}
        y = corners[1][0] + (1 if self.border is None else 2)
        x0 = corners[0][0] + (1 if self.border is None else 2)
        x1 = corners[0][1] - (1 if self.border is None else 2)
        # Commons.UIInstance.log('Paint dialog: x0={0}, x1={1}'.format(x0, x1))
        idx = 0
        for item in self.fields:
            if isinstance(item, DialogFieldButton):
                buttons.append(item)
                indices[item] = idx
            else:
                item.paint(x0, x1, y, idx == self.highlighted)
                y += 1
            idx += 1
        y = corners[1][1] - (1 if self.border is None else 2)
        if len(buttons) > 0:
            subdiv = float((x1 - x0 + 1) / len(buttons))
            width = int(subdiv)
            rem = subdiv - width
            remacc = 0
            for button in buttons:
                remacc += rem
                x1 = x0 + width
                if remacc > 0.5:
                    x1 += 1
                    remacc -= 1
                button.paint(x0, x1, y, indices[button] == self.highlighted)
                x0 = x1 + 1



"""
This module defines classes that allow simply coloring text on terminals.
"""


class Palette:
    """
    Palette base class.

    A palette is expected to be able to colorize a string in a certain terminal format based on a foreground color code,
    a background color code, and optionally, a boldness flag.
    """

    def __call__(self, code, string, background=None, bold=False):
        """
        Colorize a string.

        Parameters
        ----------
        code : object
            The color code of the foreground color. Palette implementation decides how to handle this code.
        string : str
            The string to colorize.
        background : object, optional
            The color code of the background color, or default if omitted.
        bold : bool
            Whether to bolden the string.

        Returns
        -------
        str
            The string, with the appropriate escape sequences embedded to set the color at the beginning of the string and to reset at the end.
        """
        return string


class Palette8Bit(Palette):
    """
    Palette8Bit uses the ANSI 38;5;n and 48;5;n format to access the 256 color table.
    """

    def __call__(self, code, string, background=None, bold=False):
        out = ""
        if background is not None:
            out += "\033[48;5;" + str(background) + "m"
        out += "\033["
        if bold:
            out += "1;"
        out += "38;5;" + str(code) + "m" + string + "\033[0m"
        return out


class Color:
    """
    A Color is a specific code in a palette. Color objects can be used to quickly colorize strings.

    Attributes
    ----------
    palette : awsc.termui.color.Palette
        The Palette where this color can be found.
    code : object
        The code of the foreground color associated with this color in the palette.
    background : object
        The code of the background color associated with this color in the palette. If None, the color of the background should not be changed.
    """

    def __init__(self, palette, code, background=None):
        """
        Initializes a Color object.

        Parameters
        ----------
        palette : awsc.termui.color.Palette
            The Palette where this color can be found.
        code : object
            The code of the foreground color associated with this color in the palette.
        background : object, optional
            The code of the background color associated with this color in the palette. If None, the color of the background should not be changed.
        """
        self.palette = palette
        self.code = code
        self.background = background

    def __str__(self):
        return f"{type(self.palette).__name__}:{self.code},{self.background}"

    def __repr__(self):
        return f"{type(self).__name__}({type(self.palette).__name__}, {self.code}, {self.background})"

    def __call__(self, string, bold=False):
        """
        Colorize a string with this Color.

        Parameters
        ----------
        string : str
            The string to colorize.
        bold : bool
            Whether to bolden the string.

        Returns
        -------
        str
            The colorized string ready for terminal output.
        """
        return self.palette(self.code, string, background=self.background, bold=bold)


ColorBlack = Color(Palette8Bit(), 0)
"""
Black color on a 256 color palette preset.
"""

ColorWhite = Color(Palette8Bit(), 15)
"""
White color on a 256 color palette preset.
"""

ColorWhiteOnBlack = Color(Palette8Bit(), 15, background=0)
"""
White color on black blackground on a 256 color palette preset.
"""

ColorBlackOnWhite = Color(Palette8Bit(), 0, background=15)
"""
Black color on white background on a 256 color palette preset.
"""

ColorGold = Color(Palette8Bit(), 220)
"""
Gold color on a 256 color palette preset.
"""

ColorGreen = Color(Palette8Bit(), 40)
"""
Bright green color on a 256 color palette preset.
"""

ColorDarkGreen = Color(Palette8Bit(), 70)
"""
Dark green color on a 256 color palette preset.
"""

ColorMagenta = Color(Palette8Bit(), 165)
"""
Magenta color on a 256 color palette preset.
"""

ColorBlackOnGold = Color(Palette8Bit(), 0, background=220)
"""
Black color on gold background on a 256 color palette preset.
"""

ColorBlackOnOrange = Color(Palette8Bit(), 0, background=208)
"""
Black color on orange background on a 256 color palette preset.
"""

ColorBlackOnGray = Color(Palette8Bit(), 0, background=251)
"""
Black color on gray background on a 256 color palette preset.
"""



"""
Module for EC2 instance resources.
"""
import subprocess
from pathlib import Path

from .base_control import (
    Describer,
    DialogFieldResourceListSelector,
    ListResourceDocumentCreator,
    ResourceLister,
    ResourceRefByClass,
    SelectionAttribute,
    SingleRelationLister,
    TemplateDict,
    tagged_column_generator,
)
from .common import Common, SessionAwareDialog
from .dashboard import KeyValueDashboardBlock
from .termui.control import Border
from .termui.dialog import DialogFieldCheckbox, DialogFieldText
from .termui.ui import ControlCodes


@ResourceLister.Autocommand(
    "EC2ResourceLister", "r", "View related resources", "selection"
)
class EC2RelatedLister(SingleRelationLister):
    """
    Lister control for all resources related to a single EC2 instance based on its description.
    """

    prefix = "ec2_related"
    title = "EC2 Instance related resources"

    resource_type = "ec2 instance"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Instance"
    describe_method = "describe_instances"
    describe_kwargs = TemplateDict({"InstanceIds": [SelectionAttribute("instance id")]})
    object_path = ".Reservations[0].Instances[0]"
    resource_descriptors = [
        {
            "base_path": "[.SecurityGroups[].GroupId]",
            "type": "Security Group",
            "describer": ResourceRefByClass("SGDescriber"),
        },
        {
            "base_path": '[.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value]',
            "type": "Cloudformation Stack",
            "describer": ResourceRefByClass("CFNDescriber"),
        },
        {
            "base_path": "[.VpcId]",
            "type": "VPC",
            "describer": ResourceRefByClass("VPCDescriber"),
        },
        {
            "base_path": "[.NetworkInterfaces[].NetworkInterfaceId]",
            "type": "Network Interface",
        },
        {
            "base_path": "[.BlockDeviceMappings[].Ebs.VolumeId]",
            "type": "EBS Volume",
            "describer": ResourceRefByClass("EBSDescriber"),
        },
        {
            "base_path": "[.ImageId]",
            "type": "AMI",
            "describer": ResourceRefByClass("AMIDescriber"),
        },
        {
            "base_path": "[.KeyName]",
            "type": "Keypair",
        },
        {
            "base_path": "[.InstanceType]",
            "type": "Instance Type",
            "describer": ResourceRefByClass("InstanceClassDescriber"),
        },
        {
            "base_path": "[.SubnetId]",
            "type": "Subnet",
            "describer": ResourceRefByClass("SubnetDescriber"),
        },
        {
            "base_path": "[.IamInstanceProfile.Arn]",
            "type": "Instance Profile",
            "describer": ResourceRefByClass("InstanceProfileDescriber"),
        },
    ]

    def title_info(self):
        return f"EC2: {self.parent_selection['instance id']}"


@ResourceLister.Autocommand("EBInstanceHealthLister", "v", "View instance")
class EC2Describer(Describer):
    """
    Describer control for EC2 instances.
    """

    prefix = "ec2_browser"
    title = "EC2 Instance"

    resource_type = "instance"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Instance"
    describe_method = "describe_instances"
    describe_kwarg_name = "InstanceIds"
    describe_kwarg_is_list = True
    object_path = ".Reservations[0].Instances[0]"
    default_entry_key = "instance id"


@ResourceLister.Autocommand("ASGResourceLister", "i", "View instances", "asg")
class EC2ResourceLister(ResourceLister):
    """
    Lister control for EC2 instances.
    """

    prefix = "ec2_list"
    title = "EC2 Instances"
    command_palette = ["ec2", "instance"]

    resource_type = "instance"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Instance"
    list_method = "describe_instances"
    item_path = "[.Reservations[].Instances[]]"
    columns = {
        "instance id": {
            "path": ".InstanceId",
            "size": 11,
            "weight": 0,
            "sort_weight": 1,
        },
        **tagged_column_generator("name", "name", weight=1, sort_weight=0, size=30),
        "type": {
            "path": ".InstanceType",
            "size": 10,
            "weight": 2,
        },
        "vpc": {
            "path": ".VpcId",
            "size": 15,
            "weight": 3,
        },
        "public ip": {
            "path": ".PublicIpAddress",
            "size": 15,
            "weight": 4,
        },
        "key name": {
            "path": ".KeyName",
            "size": 30,
            "weight": 5,
        },
        "state": {
            "path": ".State.Name",
            "size": 10,
            "weight": 6,
        },
        "image": {"path": ".ImageId", "hidden": True},
    }
    describe_command = EC2Describer.opener
    open_command = "r"
    primary_key = "instance id"

    def title_info(self):
        return self.title_info_data

    def __init__(self, *args, asg=None, **kwargs):
        self.title_info_data = None
        if asg is not None:
            self.list_kwargs = {
                "Filters": [
                    {
                        "Name": "tag:aws:autoscaling:groupName",
                        "Values": [asg["name"]],
                    }
                ]
            }
            self.title_info_data = f"ASG: {asg['name']}"
        super().__init__(*args, **kwargs)

    @ResourceLister.Autohotkey(ControlCodes.A, "Create AMI", True)
    def create_ami(self, _):
        """
        Hotkey callback for creating an AMI from an instance.
        """
        ListResourceDocumentCreator(
            "ec2",
            "create_image",
            None,
            initial_document={
                "BlockDeviceMappings": [
                    {
                        "DeviceName": "string",
                        "VirtualName": "string",
                        "Ebs": {
                            "DeleteOnTermination": False,
                            "Iops": 123,
                            "SnapshotId": "string",
                            "VolumeSize": 123,
                            "VolumeType": "gp2",
                            "KmsKeyId": "string",
                            "Throughput": 123,
                            "OutpostArn": "string",
                            "Encrypted": False,
                        },
                        "NoDevice": "string",
                    },
                ],
                "Description": "string",
                "Name": "string",
                "NoReboot": True,
                "TagSpecifications": [
                    {
                        "ResourceType": "capacity-reservation",
                        "Tags": [
                            {"Key": "string", "Value": "string"},
                        ],
                    },
                ],
            },
            as_json=["BlockDeviceMappings", "TagSpecifications"],
            static_fields={"InstanceId": self.selection["instance id"]},
        ).edit()

    @ResourceLister.Autohotkey(ControlCodes.S, "Start/stop instance", True)
    def stop_start_instance(self, _):
        """
        Hotkey callback for starting or stopping an instance.
        """
        if self.selection["state"] == "running":
            self.confirm_template(
                "stop_instances",
                TemplateDict(
                    {
                        "InstanceIds": [SelectionAttribute("instance id")],
                    }
                ),
                undoable=True,
                action_name="Stop",
            )(self.selection)
        elif self.selection["state"] == "stopped":
            self.confirm_template(
                "start_instances",
                TemplateDict(
                    {
                        "InstanceIds": [SelectionAttribute("instance id")],
                    }
                ),
                undoable=True,
                action_name="Start",
            )(self.selection)
        else:
            Common.Session.set_message(
                "Instance is not in running or stopped state.",
                Common.color("message_info"),
            )

    @ResourceLister.Autohotkey(ControlCodes.D, "Terminate instance", True)
    def terminate_instance(self, _):
        """
        Hotkey callback for terminating an instance.
        """
        self.confirm_template(
            "terminate_instances",
            TemplateDict(
                {
                    "InstanceIds": [SelectionAttribute("instance id")],
                }
            ),
            undoable=True,
            action_name="Terminate",
        )(self.selection)

    @ResourceLister.Autohotkey("m", "Metrics", True)
    def metrics(self, _):
        """
        Hotkey callback for listing metrics for an EC2 instance.
        """
        Common.Session.push_frame(
            ResourceRefByClass("MetricLister")(
                metric_namespace="AWS/EC2",
                dimension=("InstanceId", self.selection["instance id"]),
            )
        )

    # @ResourceLister.Autohotkey("n", "Launch new instance")
    # def new_instance(self, _):
    #    """
    #    Hotkey callback for launching a new instance.
    #    """
    #    EC2LaunchDialog(
    #        self.parent,
    #        CenterAnchor(0, 0),
    #        Dimension("80%|40", "20"),
    #        caller=self,
    #        weight=-500,
    #    )

    @ResourceLister.Autohotkey("s", "SSH", True)
    def ssh(self, _):
        """
        Hotkey callback for SSHing to an instance.
        """
        key = Common.Session.ssh_key
        keypair_call = Common.generic_api_call(
            "ec2",
            "describe_key_pairs",
            {"KeyNames": self.selection["key name"]},
            "Describe Key Pair",
            "EC2",
            subcategory="Key Pair",
            resource=self.selection["instance id"],
        )
        if keypair_call["Success"]:
            keypair_response = keypair_call["Response"]
            if len(keypair_response["KeyPairs"]) > 0:
                assoc = Common.Session.get_keypair_association(
                    keypair_response["KeyPairs"][0]["KeyPairId"]
                )
                if assoc != "":
                    key = assoc
        path = Path.home() / ".ssh" / key
        if not path.exists():
            Common.Session.set_message(
                "Selected SSH key does not exist", Common.color("message_error")
            )
            return
        if self.selection is not None:
            if self.selection["public ip"] == "":
                Common.Session.set_message(
                    "No public IP associated with instance",
                    Common.color("message_error"),
                )
            else:
                EC2SSHDialog.opener(
                    instance_entry=self.selection,
                    caller=self,
                    weight=-500,
                )


class EC2SSHDialog(SessionAwareDialog):
    """
    Custom dialog for setting SSH parameters and initiating an SSH session to an EC2 instance.
    """

    line_size = 15

    def __init__(
        self,
        *args,
        instance_entry=None,
        caller=None,
        **kwargs,
    ):
        kwargs["border"] = Border(
            Common.border("ec2_ssh", "default"),
            Common.color("ec2_ssh_modal_dialog_border", "modal_dialog_border"),
            "SSH to instance",
            Common.color(
                "ec2_ssh_modal_dialog_border_title", "modal_dialog_border_title"
            ),
            instance_entry["instance id"],
            Common.color(
                "ec2_ssh_modal_dialog_border_title_info",
                "modal_dialog_border_title_info",
            ),
        )
        super().__init__(*args, caller=caller, **kwargs)
        self.set_title_label("Connect to EC2 over SSH")
        self.instance_id = instance_entry["instance id"]
        self.ip = instance_entry["public ip"]
        def_text = (
            Common.Configuration["default_ssh_usernames"][Common.Session.ssh_key]
            if Common.Session.ssh_key in Common.Configuration["default_ssh_usernames"]
            else ""
        )
        self.use_instance_connect = DialogFieldCheckbox(
            "Use EC2 Instance Connect",
            color=Common.color(
                "ec2_ssh_modal_dialog_textfield_label", "modal_dialog_textfield_label"
            ),
            selected_color=Common.color(
                "ec2_ssh_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
        )
        self.username_textfield = DialogFieldText(
            "SSH username",
            text=def_text,
            **Common.textfield_colors("ec2_ssh"),
        )
        self.add_field(self.use_instance_connect)
        self.add_field(self.username_textfield)
        self.highlighted = 0
        self.caller = caller

    def accept_and_close(self):
        path = Path.home() / ".ssh" / Common.Session.ssh_key
        path_pub = Path.home() / ".ssh" / f"{Common.Session.ssh_key}.pub"
        if self.use_instance_connect:
            if not path_pub.exists():
                Common.Session.set_message(
                    f'Public key "{Common.Session.ssh_key}.pub" does not exist in ~/.ssh',
                    Common.color("message_error"),
                )
            with path_pub.open("r") as file:
                pubkey = file.read()
            api_kwargs = {
                "InstanceId": self.instance_id,
                "InstanceOSUser": self.username_textfield.text,
                "SSHPublicKey": pubkey,
            }
            resp = Common.generic_api_call(
                "ec2-instance-connect",
                "send_ssh_public_key",
                api_kwargs,
                "Instance Connect",
                "EC2",
                subcategory="Instance Connect",
                resource=self.instance_id,
            )
            if not resp["Success"]:
                self.close()
                return
        ssh_cmd = f"{self.username_textfield.text}@{self.ip}"
        Common.info(
            f"SSH to instance: {ssh_cmd}",
            "SSH to instance",
            "EC2",
            resource=self.instance_id,
        )
        exit_code = Common.Session.ui.unraw(
            subprocess.run,
            [
                "bash",
                "-c",
                f"ssh -o StrictHostKeyChecking=no -i {path.resolve()} {ssh_cmd}",
            ],
        )
        Common.Session.set_message(
            f"ssh exited with code {exit_code.returncode}",
            Common.color("message_info"),
        )
        self.close()

    def close(self):
        self.parent.remove_block(self)


@ResourceLister.Autocommand("EC2ResourceLister", "n", "Launch new instance", "_")
class EC2LaunchDialog(SessionAwareDialog):
    """
    Custom dialog for launching a new EC2 instance.
    """

    def __init__(self, parent, alignment, dimensions, *args, caller=None, **kwargs):
        from .resource_ami import AMIResourceLister
        from .resource_ec2_class import InstanceClassResourceLister
        from .resource_keypair import KeyPairResourceLister
        from .resource_sg import SGResourceLister
        from .resource_subnet import SubnetResourceLister

        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "Launch EC2 instance",
            Common.color("modal_dialog_border_title"),
        )
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.set_title_label("Enter AWS instance details")
        self.name_field = DialogFieldText(
            "Name:",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.name_field)
        self.instance_type_field = DialogFieldResourceListSelector(
            InstanceClassResourceLister,
            "Instance type:",
            "t2.micro",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.instance_type_field)
        self.image_field = DialogFieldResourceListSelector(
            AMIResourceLister,
            "AMI:",
            "",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.image_field)
        self.keypair_field = DialogFieldResourceListSelector(
            KeyPairResourceLister,
            "Keypair:",
            "",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
            primary_key="name",
        )
        self.add_field(self.keypair_field)
        self.subnet_field = DialogFieldResourceListSelector(
            SubnetResourceLister,
            "Subnet:",
            "",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.subnet_field)
        self.secgroup_field = DialogFieldResourceListSelector(
            SGResourceLister,
            "Security group:",
            "",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.secgroup_field)
        self.caller = caller

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = "Name cannot be blank."
            return
        if self.instance_type_field.text == "":
            self.error_label.text = "Instance type cannot be blank."
            return
        if self.image_field.text == "":
            self.error_label.text = "AMI cannot be blank."
            return
        if self.keypair_field.text == "":
            self.error_label.text = "Key pair cannot be blank."
            return

        api_kwargs = {
            "ImageId": self.image_field.text,
            "InstanceType": self.instance_type_field.text,
            "KeyName": self.keypair_field.text,
            "MinCount": 1,
            "MaxCount": 1,
            "TagSpecifications": [
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": "Name", "Value": self.name_field.text}],
                },
            ],
        }
        if self.subnet_field.text != "":
            api_kwargs["SubnetId"] = self.subnet_field.text
            if self.secgroup_field.text != "":
                secgroup_call = Common.generic_api_call(
                    "ec2",
                    "describe_security_groups",
                    {"GroupIds": self.secgroup_field.text},
                    "Describe Security Group",
                    "EC2",
                    subcategory="Security Group",
                    resource=self.secgroup_field.text,
                )
                if not secgroup_call["Success"]:
                    return
                secgroup = secgroup_call["Response"]

                subnet_call = Common.generic_api_call(
                    "ec2",
                    "describe_subnets",
                    {"SubnetIds": self.subnet_field.text},
                    "Describe Subnet",
                    "EC2",
                    subcategory="Subnet",
                    resource=self.subnet_field.text,
                )
                if not subnet_call["Success"]:
                    return
                subnet = subnet_call["Response"]

                if secgroup["VpcId"] == subnet["VpcId"]:
                    api_kwargs["SecurityGroupIds"] = [self.secgroup_field.text]
                else:
                    Common.Session.set_message(
                        "Security group and subnet belong to different VPCs.",
                        Common.color("message_error"),
                    )
                    return
        Common.generic_api_call(
            "ec2",
            "run_instances",
            api_kwargs,
            "Launch Instance",
            "EC2",
            subcategory="Instance",
            success_template="Instance {0} is launching",
        )

        super().accept_and_close()

    def close(self):
        if self.caller is not None:
            self.caller.refresh_data()
        super().close()


class EC2DashboardBlock(KeyValueDashboardBlock):
    """
    Dashboard block for EC2 health.
    """

    description = "EC2 health overview"

    labels = {
        "total": "Total instances",
        "running": "Running instances",
        "stopped": "Stopped instances",
        "terminated": "Terminated instances",
        "transitioning": "Transitioning instances",
        "impaired": "Impaired instances",
        "healthy%": "of instances healthy",
        "impaired%": "of instances impaired",
    }
    inverted = ["healthy%", "impaired%"]
    suffixes = {"healthy%": "%", "impaired%": "%"}
    thresholds = {
        "healthy%": [
            "dashboard_block_negative",
            75,
            "dashboard_block_neutral",
            100,
            "dashboard_block_positive",
        ],
        "impaired%": [
            "dashboard_block_positive",
            1,
            "dashboard_block_neutral",
            25,
            "dashboard_block_negative",
        ],
    }
    order = [
        "total",
        "running",
        "stopped",
        "terminated",
        "transitioning",
        "healthy%",
        "impaired%",
        "impaired",
    ]

    def __init__(self, *args, **kwargs):
        self.instance_health_data = {}
        super().__init__(*args, **kwargs)

    def refresh_data(self):
        resp = Common.generic_api_call(
            "ec2",
            "describe_instance_status",
            {"IncludeAllInstances": True},
            "Describe instance statuses",
            "EC2",
            subcategory="Instance",
        )
        if resp["Success"]:
            ihd = resp["Response"]["InstanceStatuses"]
            info = {}
            info["total"] = len(ihd)
            info["running"] = 0
            info["transitioning"] = 0
            info["stopped"] = 0
            info["terminated"] = 0
            info["impaired"] = 0
            for inst in ihd:
                state = inst["InstanceState"]["Name"]
                if state in ("running", "stopped", "terminated"):
                    info[state] += 1
                else:
                    info["transitioning"] += 1
                if (
                    inst["InstanceStatus"]["Status"] == "impaired"
                    or inst["SystemStatus"]["Status"] == "impaired"
                ):
                    info["impaired"] += 1
            info["impaired%"] = int((info["impaired"] / info["total"]) * 100)
            info["healthy%"] = 100 - info["impaired%"]
            with self.mutex:
                self.status = self.STATUS_READY
                self.info = info
                self.instance_health_data = ihd

        else:
            with self.mutex:
                self.status = self.STATUS_ERROR
